import uuid
from random import randint

from flask import Blueprint, Response, request
from pathlib import Path

from audio_files import *

from barix.web.auth import login_required
from barix.web.exceptions import InvalidUploadedFileError, InvalidFileExtensionError, FileIsPlayingError
from barix.web.system_info import getStorageSize
from barix.web.utils import writeToJSONFile
from barix.web.audio_files import uploadAudioFile

from utils import set_flag, clear_flag

# from ipformer import getSourcesStatus
# from scheduler import updateSchedulerPlaylists

bp = Blueprint('audio_files_API', __name__)

log = logging.getLogger('flask-backend')

p = Path(AUDIO_FILES_DIR)
if not p.exists():
    try:
        p.mkdir()
    except Exception as e:
        log.error("Error creating directory {}: {}".format(AUDIO_FILES_DIR, e))

AUDIO_FILES_ENDPOINT = '/api/rest/v1/icpaging/master/audiofiles'  # Same as in the frontend

@bp.route(AUDIO_FILES_ENDPOINT, methods=['GET'])
@login_required
def handleGetAudioFiles():
    log.info("Received GET audio files request")
    audioFiles = {}

    try:
        if len(request.args) == 0:
            audioFileTypesRequested = ['local']  # 'playlists', 'usb'
        else:
            audioFileTypesRequested = request.args
        for audiofileType in audioFileTypesRequested:
            if audiofileType == "local":
                audioFiles["local"] = getLocalAudioFiles()
            # elif audiofileType == "usb":
            #     usbAudioFiles = getUSBAudioFiles()
            #     if usbAudioFiles is not None: # means that there is a USB inserted
            #         audioFiles["usb"] = usbAudioFiles
            # elif audiofileType == "playlists":
            #     audioFiles["playlists"] = getPlaylists()
            else:
                return Response(status=404)
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        return Response(json.dumps(audioFiles), status=200)

@bp.route(AUDIO_FILES_ENDPOINT, methods=['PUT'])
@login_required
def handleSaveAudioFilesDetails():
    try:
        data = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)
    else:
        try:
            if 'local' in data.keys():
                writeToJSONFile(AUDIO_FILES_JSON_FILE, data['local'])
                generateAllStationNumbers()  # When a station number has been changed, ensures the list is updated properly
                set_flag("ringtones")  # The main app needs to be restarted, so the ringtones can be encoded correctly
            # if 'playlists' in data.keys():
            #     playlistsDict = {}
            #     for elem in data['playlists']:
            #         if "id" in elem.keys():
            #             key = elem.pop("id")
            #         else:
            #             key = uuid.uuid4().hex
            #         playlistsDict[key] = elem
            #     writeToJSONFile(PLAYLISTS_FILE_PATH, playlistsDict)
            #     updateSchedulerPlaylists(playlistsDict)
        except Exception as e:
            log.error(e)
            return Response(status=500)
        else:
            return Response(status=200)

@bp.route(AUDIO_FILES_ENDPOINT, methods=['POST'])
@login_required
def handleUploadAudioFiles():
    log.info("Successful POST request")
    freeSpace = getStorageSize()["available"]
    first = True
    try:
        errorFiles = []
        successCount = 0
        audioFilesList = request.files.getlist("audiofile")
        if len(audioFilesList) > 0:
            for uploadedFile in audioFilesList:
                try:
                    if not first:
                        # get free space size in storage
                        freeSpace = getStorageSize()["available"]
                    # Testing new implementation of audio_files.py library
                    # additional_info = {"station_number": str(randint(1,128))}
                    additional_info = {"station_number": getFreeStationNumber()}

                    n_files_uploaded = uploadAudioFile(uploadedFile, AUDIO_FILES_DIR, AUDIO_FILES_FORMAT, freeSpace,
                                    AUDIO_FILES_JSON_FILE, **additional_info)
                    

                except InvalidFileExtensionError as e:
                    log.error("Error while uploading file {}: {}".format(uploadedFile.filename, str(e)))
                    errorFiles.append("Error storing \"{}\": {}".format(uploadedFile.filename, str(e)))
                except InvalidUploadedFileError as e:
                    log.error("Error while uploading file {}: {}".format(uploadedFile.filename, str(e)))
                    errorFiles.append("Error storing \"{}\": {}".format(uploadedFile.filename, str(e)))
                except Exception as e:
                    log.error("Error while uploading file {}: {}".format(uploadedFile.filename, str(e)))
                    raise e
                else:
                    successCount += n_files_uploaded
                    refreshAvailableStationNumbers()
                    first = False
        else:
            return Response("No file selected", status=400, headers={"Content-Type" : "text/plain"})
    except Exception as e:
        log.error("Error occurred while uploading file : {}".format(str(e)))
        if str(e).startswith('413 Request Entity Too Large'):
            return Response(status=413)
        elif "No space left on device" in str(e):
            return Response(status=507)
        else:
            return Response(status=500)
    else:
        return_msg = {"success" : successCount, "errors" : errorFiles}
        set_flag("ringtones")  # The main app needs to be restarted, so the ringtones can be encoded correctly
        return Response(json.dumps(return_msg), status=200)

@bp.route(AUDIO_FILES_ENDPOINT, methods=['DELETE'])
@login_required
def handleDeleteAudioFiles():
    log.info("Successful DELETE request")
    try:
        data = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)
    else:
        try:
            log.info("Data received for deletion: %s", data)

            nAudioFilesRemoved = deleteAudioFiles(data['payload'])
            set_flag("ringtones")  # The main app needs to be restarted, so the ringtones can be encoded correctly
            return Response(json.dumps(nAudioFilesRemoved), status=200)
        except FileIsPlayingError:
            return Response(status=409)
        except Exception as e:
            log.error(e)
            return Response(status=500)

@bp.route(AUDIO_FILES_ENDPOINT + '/play', methods=['POST'])
@login_required
def handlePlayAudioFile():
    log.info("Successful /play POST request")
    try:
        data = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)
    else:
        try:
            res = playAudioFile(data['filename'])
            # set_flag("playing")
            return Response(json.dumps(res), status=200)
        except Exception as e:
            log.error(e)
            return Response(status=500)

@bp.route(AUDIO_FILES_ENDPOINT + '/stop', methods=['POST'])
@login_required
def handleStopPlayingAudioFile():
    log.info("Successful /stop POST request")
    try:
        stopPlayingAudioFile()
        # clear_flag('playing')
        return Response(status=200)
    except Exception as e:
        log.error(e)
        return Response(status=500)

@bp.route(AUDIO_FILES_ENDPOINT + '/playing', methods=['GET'])
@login_required
def handleGetPlayingFile():
    filePlaying = None
    try:
        playing_status = getPlayingStatus() # Needs implementation and testing
        if "status" in playing_status:
            for item in playing_status["status"]:
                if item["priority"] == 0 and item["type"] == "local_file":
                    # log.warning("Current state: %s",item["state"])
                    if item["state"] == "playing":
                        filePlaying = item["filename"]                
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        return Response(filePlaying, status=200)