# standard library imports
import logging
import os
from flask import Flask
from pathlib import Path
# related third party imports

# local application/library specific imports
import logger

from barix.web.restart_services import restartServicesFromFile
from barix.web.constants import SSL_CERTIFICATES_DIR
from barix.web.uci import getValueOfUci
from barix.web.certificates import generateSSLCerts

UPLOAD_FILES_DIR = "/mnt/data/uploads"
MAX_FILE_SIZE_UPLOAD = 256 * 1024 * 1024 #256Mb

CERT_PATH = os.path.join(SSL_CERTIFICATES_DIR, 'barix.pem')

def include_routes_to_frontend():
    try:
        from static.frontendRoutes import frontendRoutes
        flask_app.register_blueprint(frontendRoutes)
        #log.debug("Imported frontend routes successfully")
    except Exception as e:
        log.debug("It was not possible to import frontend routes: {}".format(e))

def include_standard_routes():
    import settings_API
    flask_app.register_blueprint(settings_API.bp)
    import auth_API
    flask_app.register_blueprint(auth_API.bp)
    import logs_API
    flask_app.register_blueprint(logs_API.bp)
    import reboot_API
    flask_app.register_blueprint(reboot_API.bp)
    import reset_to_defaults_API as defaults_API
    flask_app.register_blueprint(defaults_API.bp)
    import system_info_API
    flask_app.register_blueprint(system_info_API.bp)
    import uci_API
    flask_app.register_blueprint(uci_API.bp)
    import fw_update_API
    flask_app.register_blueprint(fw_update_API.bp)
    import ca_certificates_API
    flask_app.register_blueprint(ca_certificates_API.bp)
    import icpaging_master_API
    flask_app.register_blueprint(icpaging_master_API.bp)
    import paging_master_info_API
    flask_app.register_blueprint(paging_master_info_API.bp)
    import audio_files_API
    flask_app.register_blueprint(audio_files_API.bp)

from io import BytesIO
from tempfile import TemporaryFile, SpooledTemporaryFile
from flask.wrappers import Request

class MemoryRequest(Request):
    def _get_file_stream(self, total_content_length, content_type, filename, content_length):
        max_size = 1024 * 500
        if SpooledTemporaryFile is not None:
            return SpooledTemporaryFile(max_size=max_size, mode="rb+", dir=UPLOAD_FILES_DIR)
        elif total_content_length is None or total_content_length > max_size:
            return TemporaryFile(mode="rb+", dir=UPLOAD_FILES_DIR)
        return BytesIO()

"""
Init logger
"""
logger.configLogger('flask-backend', "/var/log/flask-backend.log", 2*1024*1024, 2)
log = logging.getLogger('flask-backend')
log.info("Starting flask backend...")

"""
Check if has services to restart
"""
restartServicesFromFile()

"""
Create non existing directories
"""
p = Path(UPLOAD_FILES_DIR)
if not p.exists():
    try:
        p.mkdir()
        #log.debug("Directory {} successfully created".format(UPLOAD_FILES_DIR))
    except Exception as e:
        log.error("Error creating directory {}: {}".format(UPLOAD_FILES_DIR, e))

"""
Init Flask application
"""
# static_url_path='' removes any preceding path from the URL (i.e. the default /static).
# static_folder='web/static' to serve any files found in the folder web/static as static files.
# template_folder='web/templates' similarly, this changes the templates folder.
flask_app = Flask(__name__, static_url_path='', static_folder='static', template_folder='static')
# secret key used to sign session cookies for protection against cookie data tampering.
# TODO: get secret key from anywhere else, like env variable, for security (?)
flask_app.secret_key = b'\xb3\xa482C\x98\x81\xa9\x8e5\xe9\x7f\xe2\xf9y/'
# UPLOAD_FOLDER is where uploaded files are stored
flask_app.config['UPLOAD_FOLDER'] = UPLOAD_FILES_DIR
# maximum size allowed for the files to upload
flask_app.config['MAX_CONTENT_LENGTH'] = MAX_FILE_SIZE_UPLOAD
# override default werkzeug formparser to allow to update files to the UPLOAD_FOLDER instead of default
flask_app.request_class = MemoryRequest

"""
Include routes
"""
include_routes_to_frontend()
include_standard_routes()

"""
Get web protocol from uci to launch application
"""

protocol = getValueOfUci('httpd', 'webserver', 'protocol')

if protocol == "https":
    if not os.path.exists(CERT_PATH):
        generateSSLCerts()

    try:
        flask_app.run(host='0.0.0.0', port=443, threaded=True, ssl_context=(CERT_PATH, CERT_PATH))
    except Exception as e:
        log.error(e)
else:
    try:
        flask_app.run(host='0.0.0.0', port=80, threaded=True)
    except Exception as e:
        log.error(e)
