import subprocess
import time
import os
import sys
import syslog
import alsaaudio

from barix.io import leds
from utils import get_logger, get_uci, execute_command

logger = get_logger(name="AlsaConfigurator")


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class AlsaConfigurator(metaclass=Singleton):
    def __init__(self,):
        self.app_uci_config = self.createAppConfigDict("application")
        # Dict with the available controls and their range in dB or their options
        # Not used at the moment
        self.available_controls = {"Line In": {"min": -4.5, "max": 6},
                                   "Line Out": [-43.5, 0],
                                   "Line Out Source": ["Stereo, Mono Differential"],
                                   "Mic1": [-4.5, 6],
                                   "Mic1 Boost": [0, 42],
                                   "Mic2": [-4.5, 6],
                                   "Mic2 Boost": [0, 42],
                                   "Mixer": [],
                                   "Mixer Reverse": [],
                                   "ADC Gain": [-4.5, 6],
                                   "DAC": [-73.08, 0]}

    @staticmethod
    def createAppConfigDict(package):
        """
            Creates config from uci to be used at startup for alsa configuration
        """
        config_dict = {"ad_gain": get_uci(package, "audio", "ad_gain"),
                       "mic_linein": get_uci(package, "audio", "mic_linein"),
                       "mic_boost": get_uci(package, "audio", "mic_boost"),
                       "mic_gain": get_uci(package, "audio", "mic_gain"),
                       "mic_boost": get_uci(package, "audio", "mic_boost"),
                       "ext_mic": get_uci(package, "audio", "ext_mic"),
                       "silent_playback": get_uci(package, "audio", "silent_playback"),
                       "spk_out": get_uci(package, "audio", "spk_out"),
                       "volume": get_uci(package, "audio", "volume"),
                       "dac_gain": get_uci(package, "audio", "dac_gain")}
        return config_dict

    def refreshConfigDict(self):
        """
            Read the UCI values from the device and update the config dict with the new values
        @return:
        """
        self.app_uci_config = self.createAppConfigDict("application")

    def setVolume(self, control: str = "Line Out", card: int = 2, volume: str = "default", capture: str = ""):
        """
            Change the volume of different controls
        @param capture: Use for input devices to activate/deactivate recording
        @param control: Type of control to be changed
        @param card: Audio card on the device
        @param volume: Volume as percentage or as dB. Make sure value and unit are concatenated. E.g. 10% or 0dB
        @return: None
        """
        commands_list = []
        if volume == "default":
            self.refreshConfigDict()
            output_device = get_uci("icpaging-master", "audio", "output_device")
            if output_device == 'plug:master1':
                # "BARIX MAIN" in dB
                volume = self.app_uci_config["volume"]
                # Line Out must be at maximum
                commands_list.append(["amixer", "-c", f"{card}", "sset", f"Line Out", "100%"])
            elif output_device == "plughw:2,0":
                # Line Out volume
                volume = "80%"
        
        if volume != "":
            commands_list.append(["amixer", "-c", f"{card}", "sset", f"{control}", "--", f"{volume}"])

        if capture != "":
            if card == 2:
                if control == "Line In" or control == "Mic1":
                    # make sure capture is activated
                    commands_list.append(["amixer", "-c", f"{card}", "sset", f"{control}", f"{capture}"])

        self.execute(commands_list)

    def setAdcGain(self):
        self.refreshConfigDict()
        self.setVolume(control='ADC Gain', card=2, volume=self.app_uci_config["ad_gain"])

    def setDacGain(self):
        self.refreshConfigDict()
        self.setVolume(control='DAC', card=2, volume=self.app_uci_config["dac_gain"])

    def setMicGain(self):
        self.refreshConfigDict()
        self.setVolume(control='Mic1', card=2, volume=self.app_uci_config["mic_gain"])

    def setMicBoost(self):
        self.refreshConfigDict()
        self.setVolume(control='Mic1 Boost', card=2, volume=self.app_uci_config["mic_boost"])

    def setMicLinein(self):
        audio_in = self.app_uci_config["mic_linein"]
        if audio_in == "micin":
            self.setVolume(control="Mic1", card=2, capture="cap")
            self.setVolume(control="Line In", card=2, capture="nocap")
            logger.info("setting Mic In")
        else:
            self.setVolume(control="Mic1", card=2, capture="nocap")
            self.setVolume(control="Line In", card=2, capture="cap")
            logger.info("setting LINE In")
    
    def setSpeakerOut(self):
        self.refreshConfigDict()
        if  "spk_out" in self.app_uci_config.keys():
            spk = self.app_uci_config["spk_out"]
            if spk == "both":
                vol = "on,on"
            elif spk == "left" or spk == "builtin":
                vol = "on,off"
            elif spk == "right" or spk == "ext":
                vol = "off,on"
            else:
                vol = "80%"

        self.setVolume(control='DAC', card=2, volume=vol)

    def setAllControls(self):
        self.setMicLinein()
        self.setVolume()
        self.setAdcGain()
        self.setMicGain()
        self.setDacGain()
        self.setSpeakerOut()
        self.setMicBoost()

    def setMute(self, control: str, card: int = 2, mute: str = 'on'):
        # amixer -c 2 sset 'Line In' toggle
        # amixer -c 2 sset 'Line In' mute/unmute
        mutable = ["Line In", "Line Out", "DAC"]
        commands_list = []

        if card == 2:
            if control in mutable:
                commands_list.append(["amixer", "-c", f"{card}", "sset", f"{control}", f"{mute}"])

        self.execute(commands_list)

    def setMicOff(self):
        self.refreshConfigDict()
        self.setVolume(control='Line In', card=2, volume=self.app_uci_config["mic_gain"], capture='nocap')

    @staticmethod
    def execute(commands_list: list):
        """
            Use this method to execute all commands in a list
        @param commands_list:
        @return:
        """

        if len(commands_list) == 0:
            logger.info('No commands to be executed')
            return
        for command in commands_list:
            try:
                result = execute_command(command)
                if result:
                    logger.info("Success executing: %s", " ".join(command))
                    pass
                else:
                    logger.info("Failed executing command: %s", " ".join(command))
                    logger.info("Trying to execute as system command")
                    try:
                        # Try to execute as a system command
                        os.system(" ".join(command))
                    except Exception as e:
                        logger.error(f"Following exception occurred while execution system command: {e}")
            except Exception as e:
                logger.warning(f"Following exception occurred: {e}")
                logger.info("Trying to execute as system command")
                try:
                    # Try to execute as a system command
                    os.system(" ".join(command))
                except Exception as e:
                    logger.error(f"Following exception occurred while execution system command: {e}")
