from pathlib import Path
from threading import Thread
from os import path, remove, system
import time

from utils import validate_directory_structure
if validate_directory_structure():
    from nextion.nextion_interface import NextionInterface
    from nextion.nextion_panel import NextionPanel
    from nextion.state_variables import ButtonState, NextionStateVariable
    from server.paging_server import PagingServer
    from shared_constants import SYS_CONFIG_FLAGS_DIR, RESET_DEFAULTS_FLAG, PAGING_MASTER_DATA
    from alsa_configurator import AlsaConfigurator
    from utils import (get_logger, get_uci, reset_defaults, restart_main_app,
                       load_alsa_settings, store_alsa_settings, NextionPanelReset)
else:
    print("Could not validate structure. Exiting")
    exit(1)


logger = get_logger("Main")
serial_logger = get_logger(name="Main-Serial", log_file='/var/log/serial_message_logger')


def set_debug_mode(panel_obj: NextionPanel, mode: str = '0', pages: str = '0'):
    import serial
    with serial.Serial(
            port=panel_obj.serial_port,
            baudrate=panel_obj.baud_rate,
            timeout=panel_obj.timeout
    ) as ser:
        # Show debugging fields on the screen
        TERMINATING_CHARACTER = b"\xff\xff\xff"
        mode_message = f"debug={mode}"
        ser.write(mode_message.encode() + TERMINATING_CHARACTER)
        pages_message = f"key_ext={pages}"
        ser.write(pages_message.encode() + TERMINATING_CHARACTER)
        serial_logger.info(f"\nMode message sent: {mode_message},\npages message sent: {pages_message}")


def is_half_duplex():
    if get_uci("icpaging-master", "audio", "duplex_mode") == 'half':
        return True
    return False


if __name__ == "__main__":

    load_alsa_settings()
    defaults = path.join(SYS_CONFIG_FLAGS_DIR, RESET_DEFAULTS_FLAG)
    if path.exists(defaults):
        remove(defaults)  # Clear the flag immediately
        reset_defaults()

    # Make sure the icpaging-master dir exists
    Path(PAGING_MASTER_DATA).mkdir(parents=True, exist_ok=True)

    alsa_configurator = AlsaConfigurator()
    alsa_configurator.setAllControls()

    nextion_interface = NextionInterface(
        nextion_state=NextionStateVariable(),
        buttons_state=ButtonState(),
        half_duplex=is_half_duplex(),
    )

    n_pages = get_uci("icpaging-master", "main", "panel_pages")
    panel = NextionPanel(
        nextion_interface=nextion_interface,
        baud_rate=115200,
        serial_port="/dev/ttyS1",
        number_pages=int(n_pages),
    )
    # server = PagingServer(nextion_interface)
    server = PagingServer(panel, nextion_interface)
    panel_thread = Thread(target=panel.run, daemon=True)
    try:
        panel_thread.setName("Thread-NextionPanel")
        panel_thread.start()
        set_debug_mode(panel_obj=panel,
                       mode=get_uci("icpaging-master", "main", "panel_debug"),
                       pages=n_pages)
        server.run()
    except NextionPanelReset:
        store_alsa_settings()
        logger.warning("Restarting the app after a panel reset")
        server.stop()
        panel.stop()
        restart_main_app(system_func=True)

    except KeyboardInterrupt:
        store_alsa_settings()
        server.stop()
        panel.stop()
    except Exception as e:
        print(f"[main.py] Exiting: {e}")
        server.stop()
        panel.stop()
        # Restart the app to avoid unusable state
        print(f"[main.py] Restarting the main app...")
        restart_main_app(system_func=True)
