import logging
import logging.config
import socket
import time
from datetime import datetime

from BARP.barp import BARP

from master.constants import CONSTANTS
from master.services.call_audio_service import CallAudioService
from master.services.page_audio_service import PageAudioService
from master.utils import is_multicast, is_broadcast, master_uci_to_config_dict
from utils import get_logger

# configure logger
logger = get_logger("PagingMaster")


class PagingMaster:
    """
    Class defining basic IC paging master attributes.
    """

    # Time in seconds after which a client is declared stale
    # during call request before call accept
    CALLING_CLIENT_TIMEOUT = 5

    def __init__(self):
        # configure logger
        # logging.config.fileConfig("logging.conf")
        # self.logger = logging.getLogger("PagingMaster")

        try:
            config = master_uci_to_config_dict()
        except Exception as e:
            logger.error("PagingMaster initialization failed : {e}")
            raise Exception("PagingMaster initialization failed")
        # Variable initializations.
        self.barp = BARP(config["id"])

        self._device_ip = config["device_ip_settings"]["inet"]
        self._device_netmask = config["device_ip_settings"]["netmask"]
        self._broadcast = config["device_ip_settings"]["broadcast"]

        self._barp_id = config["id"]
        self.header_kwargs = config["barp_header_kwargs"]
        self.header_kwargs["sender_id"] = config["id"]
        self.header_kwargs["receiver_id"] = 0
        self.__relay_activation_time = config["relay_activation_time"]
        self._ports = config["ports"]
        self._encoding_type = config["encoding_type"]
        self._multicast_group_address = config["multicast_group_address"]
        self._page_group_request_address = "0.0.0.0"  # Not tested
        self._page_group_request_port = 0  # Not tested
        self._current_unicast_port = 0
        self._state = 0  # starts with IDLE state.
        self._barp_ip = config["ip"]
        self._exit = False
        self._in_call = False
        self.__client_number_in_call = None
        self._end_message_transmission = False
        self.rejected_call = {}
        self.calling_client = {}
        self.active_call_client = {}
        self.clients = {}
        self.is_listen_mode = False  # only relevant for half duplex
        self.client_lost = False
        self.capture_disabled = False

        # Constants.
        self.PAGING_REQUEST_TIME_INTERVAL = CONSTANTS.paging_request_time_interval
        self.STATUS_REPORT_TIME_INTERVAL = CONSTANTS.status_report_time_interval

        # Audio services.
        self.paging_audio_service = PageAudioService()
        self.call_audio_service = CallAudioService()

        # Socket initializations.
        self.listen_status_socket = None
        self.status_socket = None
        self.control_socket = None
        self.call_socket = None  # Initialised right before they are used. Closed when not needed
        self.audio_socket = None  # Initialised right before they are used. Closed when not needed
        self.initialize_sockets()

    @property
    def unicast_port(self):
        return self._current_unicast_port

    @property
    def barp_id(self):
        return self._barp_id

    @property
    def end_message_transmission(self):
        return self._end_message_transmission

    @property
    def exit_flag(self):
        return self._exit

    @property
    def state(self):
        """
        Possibilites:
        0(IDLE) and 130(Sending the message)
        """
        return self._state

    @property
    def control_port(self):
        """
        Port used for sending the control messages.
        """
        return self._ports["control"]

    @property
    def audio_port(self):
        """
        Port used for sending the audio message/stream/call
        """
        return self._ports["audio"]

    @property
    def status_port(self):
        """
        Port used for sending the  status message
        """
        return self._ports["status"]

    @property
    def master_audio_port(self):
        return self._ports["master_audio"]

    @property
    def encoding_type(self):
        """
        Encoding type to be used for the audio message.
        """
        return self._encoding_type

    @property
    def multicast_group(self):
        """
        Multicast group address. To be used while sending the paging message.
        """
        if is_multicast(self._multicast_group_address):
            logger.warning(f"Multicast address: {self._multicast_group_address}")
            return self._multicast_group_address
        else:
            return self.barp_ip

    @property
    def page_group_request_address(self):
        return self._page_group_request_address

    @property
    def page_group_request_port(self):
        return self._page_group_request_port

    @property
    def barp_ip(self):
        if is_broadcast(self._barp_ip, self._broadcast):
            return self._broadcast
        else:
            return self._barp_ip

    @property
    def in_call(self):
        """Returns true if in call, else false."""
        return self._in_call

    @property
    def client_number_in_call(self):
        """Returns the client number that is currently in call"""
        return self.__client_number_in_call


    @staticmethod
    def _generate_random_unicast_port(start: int, end: int):
        """
            Generate a random Unicast port number, to be used during current communication
        @param start: Start value of the unicast port range
        @param end: End value of the Unicast port range
        @return: Unicast port number
        """
        import random
        return random.randint(start, end)

    def _open_audio_socket(self, socket_type="", unicast_address=""):
        """
            Currently only matters if the socket should be UNICAST or not. Therefore, if no parameter is provided
            for socket_type, a socket will be opened based on the current device IP configuration and MULTICAST or
            BROADCAST socket will be created
        @param socket_type: leave empty for MULTICAST/BROADCAST, otherwise give UNICAST as a parameter
        @param unicast_address: if UNICAST socket is needed, provide the corresponding IP address of the client
        @return: socket handle
        """

        socket_handle = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

        if socket_type != "UNICAST":
            if is_multicast(self._multicast_group_address):
                socket_handle.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
                self._page_group_request_address = self._multicast_group_address
                self._page_group_request_port = self.master_audio_port

                self._page_group_request_address = self._multicast_group_address
                self._page_group_request_port = self.master_audio_port

                socket_handle.bind((self.page_group_request_address, self.page_group_request_port))

                logger.info("Audio Multicast Group IP is multicast, address: %s, port: %s", self.page_group_request_address,
                            self.page_group_request_port)

            else:
                logger.info("Audio Multicast Group IP is NOT multicast.")

                self._page_group_request_address = "0.0.0.0"
                # self._page_group_request_port = 0
                self._page_group_request_port = self.audio_port

                if is_multicast(self.barp_ip):
                    logger.info("Barp IP is multicast")
                    socket_handle.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
                    socket_handle.bind((self.barp_ip, self.audio_port))

                elif is_broadcast(self.barp_ip, self._broadcast):
                    logger.info("Barp IP is broadcast")
                    socket_handle.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                    socket_handle.bind((self.barp_ip, self.audio_port))
                else:
                    logger.warning("{} is NOT multicast OR broadcast. Defaulting to broadcast".format(self.barp_ip))
                    # TODO: Needs to be changed to reflect the actual barp_ip. Find a replacement for socket.SO_BROADCAST

                    socket_handle.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                    socket_handle.bind(("", self.audio_port))
        else:
            # logger.warning("UNICAST socket needs to be configured")

            self._current_unicast_port = self._generate_random_unicast_port(3000, 3500)
            logger.info("Using address: %s, port: %s", unicast_address, self._current_unicast_port)
        # logger.info("Socket opened: %s", socket_handle)
        return socket_handle

    def initialize_sockets(self):
        """
        Initializes the socket for sending the status of master (status_socket),
        receiving status message from other devices (listen_status_socket),
        sending control message (control_socket) and sending audio (audio socket).
        """
        self.listen_status_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.status_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self.control_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

        if is_multicast(self.barp_ip):
            self.control_socket.setsockopt(socket.SOL_SOCKET, socket.IP_MULTICAST_TTL, 1)
            self.status_socket.setsockopt(socket.SOL_SOCKET, socket.IP_MULTICAST_TTL, 1)
            self.listen_status_socket.setsockopt(socket.SOL_SOCKET, socket.IP_MULTICAST_TTL, 1)

            self.listen_status_socket.bind((self.barp_ip, self.status_port))
        else:
            self.control_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.status_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.listen_status_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

            self.listen_status_socket.bind(("", self.status_port))

        self.control_socket.bind(("", self.control_port))

    def exit(self):
        self._exit = True

    def send_status_message(self):
        """
        To be started as a parallel thread process, with a lock on state.
        Prepares a state message for the paging master.
        In this POC, only state 0 (Idle) and 130 (paging message)
        for other possible states please refer to the barp.pdf doc
        """
        logger.debug("START master status message")
        while not self.exit_flag:
            message_kwargs = {
                "state": self.state,
                "priority": 0,
                "sender_id": self.barp.my_id,
                "volume": 50,
                "config_bgm_channel": 0,
            }
            status = self.barp.encode(
                message_type=128, message_kwargs=message_kwargs, **self.header_kwargs
            )
            self.status_socket.sendto(status, (self.barp_ip, self.status_port))
            time.sleep(self.STATUS_REPORT_TIME_INTERVAL)
        logger.debug("STOP master status message")

    def send_page_group_request(self, receiver_id, priority=0, audio_file_path=None):
        """
        Prepares and sends state message 3 to be sent to the client
        before starting the stream.
        Args:
            receiver_id (int)
            priority (int)
            audio_file_path (str)
        """
        self._state = 130

        self.audio_socket = self._open_audio_socket(socket_type="")  # Needs to be closed on exit

        if is_multicast(self._multicast_group_address):
            self.paging_audio_service.update_config(
                audio_socket=self.audio_socket,
                address=(self.page_group_request_address, self.page_group_request_port),
            )
            logger.info("BARP IP: {}".format(self.barp_ip))
        else:
            self.paging_audio_service.update_config(
                audio_socket=self.audio_socket,
                address=(self.barp_ip, self.audio_port),
            )
            logger.info("BARP IP: {}".format(self.barp_ip))
        self.capture_disabled = False

        if audio_file_path is not None:
            logger.info(f"Setting file to play [{audio_file_path}]...")
            self.paging_audio_service.set_file_to_play(audio_file_path)
        self.paging_audio_service.run()  # This pages the audio
        logger.info(f"START Paging request message")

        logger.info("page_group_request_port: {}".format(self.page_group_request_port))
        logger.info("page_group_request_address: {}".format(self.page_group_request_address))

        logger.info(f"[control_socket] Sending to ip: {self.barp_ip}")
        logger.info(f"[control_socket] Sending to port: {self.control_port}")
        logger.info(f"Using priority: {priority}")

        while self._state == 130:
            # Signal the paging server that the file playing has finished
            # logger.info("Paging audio service state: %s", self.paging_audio_service.paging_state)
            if not self.paging_audio_service.paging_state:
                self._end_message_transmission = True  # Set back to false in send_end_group_request
                break

            message_kwargs = {
                "port": self.page_group_request_port,
                "priority": priority,
                "audio_encdoing_type": self.encoding_type,
                "multicast_stream_group": self.page_group_request_address,
                "receiver_id": receiver_id,
            }
            request_msg = self.barp.encode(
                message_type=3, message_kwargs=message_kwargs, **self.header_kwargs
            )

            self.control_socket.sendto(request_msg, (self.barp_ip, self.control_port))
            # send the audio
            if self.paging_audio_service.is_alsa_capture_lost:
                self.capture_disabled = True
                self._state = 0
                break
            time.sleep(self.PAGING_REQUEST_TIME_INTERVAL)
        logger.info(f"STOP Paging request message")

    def end_paging(self, receiver_id):
        """
            Prepares to end the paging and sends the end group request, MSG 8.
            The stream should stop after this message is sent.
        @param receiver_id: List of the selected clients
        @return:
        """
        logger.debug("SEND Paging end request")
        self._end_message_transmission = False
        self._state = 0
        self.paging_audio_service.exit()
        self.send_end_group_request(port=self.page_group_request_port, receiver_id=receiver_id, ip=self.barp_ip)

        # Close the socket before end
        self.audio_socket.close()
        logger.info("Audio socket closed!")

        logger.debug("SENT Paging end request")

    def listen_client_status(self):
        self.listen_status_socket.settimeout(0.5)
        logger.debug("START listening to client status")
        while not self.exit_flag:
            try:
                data, address = self.listen_status_socket.recvfrom(4096)
                msg = self.barp.decode(data)
                if msg[0]["message_type"] == 129:
                    self.update_calling_client(msg, address)
                elif msg[0]["message_type"] == 128:
                    self.update_available_clients(msg)
            except socket.timeout:
                time.sleep(0.5)
            except Exception as e:
                logger.warning(f"Client status error {e}")
        logger.debug("STOP listening to client status")

    def purge_stale_calling_client(self):
        if self.calling_client:
            is_calling_client_timedout = (
                datetime.now() - self.calling_client["time"]
            ).seconds > PagingMaster.CALLING_CLIENT_TIMEOUT

            is_calling_client_rejected_few_second_ago = (
                self.rejected_call
                and (self.rejected_call["ip"] == self.calling_client["ip"])
                and (datetime.now() - self.rejected_call["time"]).seconds
                < PagingMaster.CALLING_CLIENT_TIMEOUT
            )
            if is_calling_client_timedout or is_calling_client_rejected_few_second_ago:
                self.calling_client = {}

    def update_available_clients(self, msg):
        header, body = msg[0], msg[1]
        self.clients[header["sender_id"]] = {
            "time": datetime.now(),
            "status": body["state"],
        }

    def update_calling_client(self, msg, address):

        header = msg[0]
        if header["receiver_id"] != self.barp_id:
            return

        # Checking if the client was rejected in last few seconds.
        # To take care of stale messages on the network.
        if (
            self.rejected_call
            and self.rejected_call["ip"] == address[0]
            and (datetime.now() - self.rejected_call["time"]).seconds < 1
        ):
            logger.debug(f"Client rejected a second ago")
            return

        if len(self.calling_client) == 0:
            self.calling_client = {
                "id": header["sender_id"],
                "ip": address[0],
                "message": msg,
                "time": datetime.now(),
            }
        else:
            # Updating the client if the previous client has turned stale.
            if address[0] != self.calling_client["ip"]:
                if (
                    datetime.now() - self.calling_client["time"]
                ).seconds > PagingMaster.CALLING_CLIENT_TIMEOUT:
                    self.calling_client = {
                        "id": header["sender_id"],
                        "ip": address[0],
                        "message": msg,
                        "time": datetime.now(),
                    }
                else:
                    logger.warning(
                        f"Call from {self.calling_client['ip']} is still under consideration"
                        f"Call from {header['sender_id']} ignored!"
                    )
            else:
                self.calling_client["time"] = datetime.now()

    def full_duplex_call(self,  mode="NORMAL", filepath=""):
        client_ip = self.active_call_client["ip"]
        self.__client_number_in_call = self.active_call_client["id"]
        logger.info(f"START full duplex call to {client_ip}")

        self.call_audio_service.update_config(
            client_ip=client_ip,
            recv_port=self.unicast_port,  # self.master_audio_port
            send_port=self.audio_port,
            full_duplex=True,
            mode=mode
        )

        # logger.warning("Client: %s, recv_port: %s, send_port: %s, mode: %s",
        #                client_ip, self.unicast_port, self.audio_port, mode)

        # start the call audio service.
        self.call_audio_service.run(filepath)  # Filepath should only be set if in UNATTENDED mode

        self.client_lost = False
        self.capture_disabled = False
        # Request playback of a stream
        while self._state == 130:
            # Signal the paging server that the file playing has finished
            if not self.call_audio_service.running_state:
                self._end_message_transmission = True

            logger.debug("Request stream playback")
            self.send_start_stream_to_client(client_ip)

            if (
                self.call_audio_service.is_connection_lost
                or self.call_audio_service.is_alsa_capture_lost
            ):
                self.client_lost = self.call_audio_service.is_connection_lost
                self.capture_disabled = self.call_audio_service.is_alsa_capture_lost
                self._state = 0
                self.call_audio_service.exit()
                break
            time.sleep(1)  # As per BARP protocol
        # self.send_stop_stream_to_client(client_ip)  # Done in the close_call()
        logger.info("END full duplex call.")
        self._in_call = False

    def half_duplex_talk_mode(self):
        if self.in_call:
            logger.info("Talk mode.")
            self.call_audio_service.talk = True
            self.call_audio_service.listen = False
            # Ask client to stop sending message.
            self.send_stop_stream_to_client(self.active_call_client["ip"])
            self.is_listen_mode = False

    def half_duplex_listen_mode(self):
        if self.in_call:
            logger.info("Listen mode.")
            self.call_audio_service.talk = False
            self.call_audio_service.listen = True
            self.is_listen_mode = True

    def half_duplex_call(self, mode="NORMAL", filepath=""):
        client_ip = self.active_call_client["ip"]
        self.__client_number_in_call = self.active_call_client["id"]
        logger.info(f"START half duplex call to {client_ip}, mode: {mode}")
        self.call_audio_service.update_config(
            client_ip=client_ip,
            recv_port=self.unicast_port,  # self.master_audio_port
            send_port=self.audio_port,
            full_duplex=False,
            mode=mode
        )
        self.client_lost = False
        self.capture_disabled = False
        self.call_audio_service.run(filepath)
        # # If the mode is UNATTENDED then we need to start in talk
        self.half_duplex_talk_mode() if mode == "UNATTENDED" else self.half_duplex_listen_mode()
        # Request playback of a stream
        while self.state == 130:
            # Signal the paging server that the file playing has finished
            if not self.call_audio_service.running_state:
                self._end_message_transmission = True

            while self.call_audio_service.listen:
                # Make client send audio message.
                self.send_start_stream_to_client(client_ip)
                time.sleep(1)  # * self.call_audio_service.listen

                if (
                    self.call_audio_service.is_connection_lost
                    or self.call_audio_service.is_alsa_capture_lost
                ):
                    self.client_lost = self.call_audio_service.is_connection_lost
                    self.capture_disabled = self.call_audio_service.is_alsa_capture_lost
                    self._state = 0
                    self.call_audio_service.exit()
                    break
            time.sleep(0.001)

        self.send_stop_stream_to_client(client_ip)
        logger.info("END half duplex call.")
        self._in_call = False

    def set_state(self, new_state):
        self._state = new_state

    def send_stop_stream_to_client(self, calling_client_ip):
        logger.info(f"SEND stop stream message to client {calling_client_ip}")
        request_msg = self.barp.encode(
            message_type=1, message_kwargs={}, **self.header_kwargs
        )
        self.call_socket.sendto(request_msg, (calling_client_ip, self.control_port))

    def send_start_stream_to_client(self, calling_client_ip):
        # logger.info("Sending stream request to client in call")
        request_msg = self.barp.encode(
            message_type=2,
            message_kwargs={
                "audio_port": self.unicast_port,  # self.master_audio_port
            },
            **self.header_kwargs,
        )
        self.call_socket.sendto(request_msg, (calling_client_ip, self.control_port))

    def send_end_group_request(self, ip, port, receiver_id=list):
        # Used to have self.audio_port
        message_kwargs = {"port": port, "receiver_id": receiver_id}
        end_page_request = self.barp.encode(
            message_type=8, message_kwargs=message_kwargs, **self.header_kwargs
        )
        self.control_socket.sendto(end_page_request, (ip, self.control_port))

    def send_call_accept_message(self):
        # To accept a call master sends message_type 3 as unicast
        # Page request
        logger.debug("SEND Call accept message")
        self._in_call = True
        self.active_call_client = self.calling_client.copy()
        self.calling_client = {}
        calling_client_ip = self.active_call_client["ip"]
        self._state = 130

        # Needs to be closed when the call is finished
        self.call_socket = self._open_audio_socket(
            socket_type="UNICAST",
            unicast_address=calling_client_ip
        )

        message_kwargs = {
            "port": self.unicast_port,  # self.master_audio_port
            "priority": 0,
            "audio_encdoing_type": self.encoding_type,
            "multicast_stream_group": self.multicast_group,
            "receiver_id": self.active_call_client["id"],
        }
        request_msg = self.barp.encode(
            message_type=3, message_kwargs=message_kwargs, **self.header_kwargs
        )

        # unicast
        self.call_socket.sendto(request_msg, (calling_client_ip, self.control_port))
        logger.debug("SENT Call accept message")

    def close_call(self):
        self._state = 0
        self.send_stop_stream_to_client(self.active_call_client["ip"])
        self.call_audio_service.exit()
        self._end_message_transmission = False
        self.__client_number_in_call = None
        self.call_socket.close()
        self._in_call = False
        logger.info("Call closed")

    def reject_call(self, reason: int = 2):
        """
            Reject an incoming call.
        @param reason:
                0: busy with other party
                1: rejected by the user
                2: not available, call next one in list (call forwarding)
        """

        if len(self.calling_client) != 0:
            client_ip = self.calling_client["ip"]
            client_id = self.calling_client["id"]
            self.rejected_call = {
                "ip": client_ip,
                "time": datetime.now(),
            }
            message_kwargs = {"reason": reason}
            reject_message = self.barp.encode(
                message_type=7,
                flags=12,
                sender_id=self.header_kwargs["sender_id"],  # Used to be 4
                receiver_id=client_id,
                sequence_number=0,
                message_kwargs=message_kwargs,
            )
            # This message is unicast
            self.control_socket.sendto(reject_message, (client_ip, self.control_port))
            logger.info(f"Call from {client_ip} rejected")

    def send_relay_message(self, receiver_id, relay_number):
        message_kwargs = {
            "relay_number": relay_number,
            "activation_time": self.__relay_activation_time,
            "receiver_id": receiver_id,
        }
        header_kwargs = self.header_kwargs.copy()
        header_kwargs["receiver_id"] = receiver_id
        relay_control_message = self.barp.encode(
            message_type=4, message_kwargs=message_kwargs, **self.header_kwargs
        )
        self.control_socket.sendto(
            relay_control_message, (self.barp_ip, self.control_port)
        )
        logger.info(f"Relay control message send to client {receiver_id}, relay {relay_number}")
