import socket
import time
from threading import Thread

import alsaaudio
import audioop
from master.utils import audio_uci_to_config_dict
from utils import get_logger
from queue import Queue
from server.services.sound_level_updater import SoundLevelUpdater, get_sound_level

logger = get_logger("CallAudioService")


class CallAudioService:
    """Class to control the audio(capture/playback) aspect of incoming call to
    paging master. The half duplex and full duplex mode are supported by toggling "talk" and "listen" flags.

    methods:
        update_config : To update the config required for starting a call.
        run  : Start the service.
        exit : Stops the service.
    For usage see paging_master's class definition.
    """

    def __init__(self):
        # configure logger
        # logging.config.fileConfig("logging.conf")
        # self.logger = logging.getLogger("CallAudioService")
        try:
            config = audio_uci_to_config_dict()
        except Exception as e:
            logger.error(f"CallAudioService initialization failed!: {e}")
            raise Exception("CallAudioService initialization failed!")
        self.__exit_flag = False
        self.__running_state_flag = False
        self.__audio_format = config["audio_format"]
        if self.__audio_format == "alaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_A_LAW
        elif self.__audio_format == "ulaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_MU_LAW
        else:
            raise Exception(
                "Audio codec not recognized by Call audio service, valid option 'ulaw' and 'alaw'."
            )
        self.__play_device = config["output_device"]
        self.__record_device = config["input_device"]
        self.__frequency = config["sample_rate"]
        self.__period_size = 5000  # config["sample_rate"]
        self.__num_channel = config["num_channels"]
        self.__audio_recv_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        self.__audio_send_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )

        # To be obtained from the paging master.
        self.__recv_port = None
        self.__send_port = None
        self.__client_ip = None
        self.__client_ip = None
        self.__mode = None  # "NORMAL" or "UNATTENDED"

        # These variables are interacted with pmaster.
        self.full_duplex = False
        self.talk = True
        self.listen = True
        self.is_connection_lost = False
        self.is_alsa_capture_lost = False

        # Init SoundLevelUpdater objects
        self.__input_sound_level_updater = SoundLevelUpdater()
        self.__output_sound_level_updater = SoundLevelUpdater()

        # Thread handles
        self.__audio_sending_thread = None
        self.__audio_receiving_thread = None
        self.__input_sound_level_updater_thread = None
        self.__output_sound_level_updater_thread = None

        # Message que to be used for receiving intercom data
        self.__Q_SIZE = 100  # Don't reassign this value
        self._listen_message_q = None
        self._listen_message_q_thread = None
        self._listen_message_q_buff_msg = 0
        self.packages_read = 0

    # # TODO: Not implemented and tested. Need to use the same socket object from paging master
    # def update_socket_config(self, audio_socket, address):
    #     self.__address = address  # Address contains both address and port
    #     self.__audio_send_socket = audio_socket

    def update_config(self, client_ip: str, recv_port: int, send_port: int, full_duplex: bool, mode: str):

        self.__client_ip = client_ip
        self.__recv_port = recv_port
        self.__send_port = send_port
        self.full_duplex = full_duplex
        self.__mode = mode
        if full_duplex:
            self.talk = True
            self.listen = True
            if self.__mode == "UNATTENDED":
                self.listen = False
        else:
            # Half duplex starts in Listen mode
            logger.info("Half duplex mode")
            self.talk = False
            self.listen = True
            if self.__mode == "UNATTENDED":
                logger.info("Call in UNATTENDED mode")
                self.talk = True
                self.listen = False
        # sockets to be initialized again.
        self.__audio_recv_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        self.__audio_send_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        self.is_connection_lost = False
        self.is_alsa_capture_lost = False

    @property
    def running_state(self):
        return self.__running_state_flag

    def run(self, filepath=""):

        # logger.warning("Client: %s, recv_port: %s, send_port: %s, mode: %s",
        #                self.__client_ip, self.__recv_port, self.__send_port, self.__mode)

        if self.__audio_sending_thread is None:
            if self.__mode == "NORMAL":
                self.__audio_sending_thread = Thread(target=self._record_and_send_audio)
                self.__audio_sending_thread.daemon = True
                self.__audio_sending_thread.setName("Thread-SendingAudio")
                self.__audio_sending_thread.start()
            elif self.__mode == "UNATTENDED":
                self.__audio_sending_thread = Thread(target=self._reply_with_away_message,
                                                     kwargs={"audio_file_path": filepath})
                self.__audio_sending_thread.daemon = True
                self.__audio_sending_thread.setName("Thread-SendingAudio")
                self.__audio_sending_thread.start()
            else:
                logger.warning("Mode not supported")
        else:
            logger.warning(
                "Attempt to start CallAudioService (send audio). Already running. Ignored!"
            )

        if self.__mode != "UNATTENDED":
            if self.__audio_receiving_thread is None:
                self.__audio_receiving_thread = Thread(target=self._play_audio)
                self.__audio_receiving_thread.daemon = True
                self.__audio_receiving_thread.setName("Thread-ReceivingAudio")
                self.__audio_receiving_thread.start()
            else:
                logger.warning(
                    "Attempt to start CallAudioService (receive audio). Already running. Ignored!"
                )

        if self.__input_sound_level_updater_thread is None:
            self.__input_sound_level_updater_thread = Thread(target=self.__input_sound_level_updater.run)
            self.__input_sound_level_updater_thread.daemon = True
            self.__input_sound_level_updater_thread.setName("Thread-InputSoundLvlUpdater")
            self.__input_sound_level_updater_thread.start()

        if self.__output_sound_level_updater_thread is None:
            self.__output_sound_level_updater_thread = Thread(target=self.__output_sound_level_updater.run)
            self.__output_sound_level_updater_thread.daemon = True
            self.__output_sound_level_updater_thread.setName("Thread-OutputSoundLvlUpdater")
            self.__output_sound_level_updater_thread.start()

    def exit(self):
        self.talk = False
        self.listen = False
        self.__exit_flag = True
        if self.__audio_sending_thread is not None:
            self.__audio_sending_thread.join()
            self.__audio_sending_thread = None
            # logger.info("--->>> __audio_sending_thread joined() <<<---")
        if self.__audio_receiving_thread is not None:
            self.__audio_receiving_thread.join()
            self.__audio_receiving_thread = None
            # logger.info("--->>> __audio_receiving_thread joined() <<<---")
        if self.__input_sound_level_updater_thread is not None:
            self.__input_sound_level_updater.exit()
            self.__input_sound_level_updater_thread.join()
            self.__input_sound_level_updater_thread = None
            # logger.info("--->>> __input_sound_level_updater_thread joined() <<<---")
        if self.__output_sound_level_updater_thread is not None:
            self.__output_sound_level_updater.exit()
            self.__output_sound_level_updater_thread.join()
            self.__output_sound_level_updater_thread = None
            # logger.info("--->>> __output_sound_level_updater_thread joined() <<<---")

        logger.info("Successful EXIT from Call Audio Service")
        self.__exit_flag = False

    def _get_audio_device(self, is_input=True):
        try:
            if is_input:
                audio_device = alsaaudio.PCM(
                    alsaaudio.PCM_CAPTURE,
                    alsaaudio.PCM_NONBLOCK,
                    device=self.__record_device,
                )
                audio_device.setperiodsize(self.__period_size)

            else:
                audio_device = alsaaudio.PCM(
                    alsaaudio.PCM_PLAYBACK,
                    alsaaudio.PCM_NONBLOCK,  # PCM_NORMAL
                    device=self.__play_device,  # config["output_device"]
                )
                audio_device.setperiodsize(256)  # self.__period_size

            audio_device.setrate(self.__frequency)
            audio_device.setchannels(self.__num_channel)
            audio_device.setformat(self.__audio_codec)

            return audio_device
        except Exception as e:
            logger.error(f"Failure: Audio device inaccessible: {e}")
            raise Exception("CallAudioService Failed!")

    def _record_and_send_audio(self):
        self.__running_state_flag = True
        logger.info(f"Start audio capture and unicast to {self.__client_ip} .")
        audio_input_dev = self._get_audio_device(is_input=True)
        self.is_alsa_capture_lost = False
        update_s_level_delay = 0
        start_at_zero = True
        with self.__audio_send_socket:
            while not self.__exit_flag:
                num_no_capture = 0
                prev_capture = True
                if start_at_zero:
                    start_at_zero = False
                    self.__input_sound_level_updater.set_new_sound_level(direction="input",
                                                                         new_value=0)
                while self.talk:
                    if not start_at_zero:
                        start_at_zero = True
                    l, data = audio_input_dev.read()
                    if l:
                        self.__audio_send_socket.sendto(
                            data, (self.__client_ip, self.__send_port)
                        )
                        if self.__input_sound_level_updater_thread is not None:
                            if update_s_level_delay % 200 == 0:
                                sound_level = audioop.rms(data, 2)
                                self.__input_sound_level_updater.set_new_sound_level(direction="input",
                                                                                     new_value=sound_level)
                            update_s_level_delay += 1
                        prev_capture = True
                    else:
                        if prev_capture:
                            num_no_capture = 0
                        else:
                            num_no_capture += 1
                        prev_capture = False
                    if num_no_capture > 1000:
                        logger.warning(
                            f"No data received from alsa capture, for over {num_no_capture} frames"
                        )
                        self.is_alsa_capture_lost = True
                    time.sleep(0.001)
                time.sleep(0.001)  # NOTE: Prevents excessive CPU usage in case talk is False (e.g. in half duplex)
        logger.info(f"Stop audio capture and unicast to {self.__client_ip}")
        self.__running_state_flag = False

    def __configure_sending_params(self):
        """
            Calculates the data block size and delay between sending each block based on the device configuration.
        @return: data_block_size, delay, burst_pkgs
        """
        burst_pkgs = 5
        if self.__frequency == 8000:
            data_block_size = 960
            delay = 0.12
            # delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000)) - 0.001
        elif self.__frequency == 24000:
            # data_block_size = 512  # 1280 , 720
            # delay = 0.022
            # delay = float("{0:.4f}".format(data_block_size / 2 / self.__frequency)) - 0.001
            # TODO: Works okay with a bit of noise. Find how to have clear sound
            data_block_size = 720
            delay = 0.0305  # 40ms orig
            burst_pkgs = 3

        else:
            data_block_size = 1280
            delay = 0.079
            logger.warning("Sample rate not supported. Using default values")
        return data_block_size, delay, burst_pkgs

    def __transmit_data(self, audio_file: bytes, burst_pkgs: int, data_block_size: int, delay: float):
        """
            This function sends the decoded audio data over to the client
        @param audio_file: Decoded audio data in bytes
        @param burst_pkgs: Number of fast packages at the beginning of transmission
        @param data_block_size: Data block size, calculated based on the device configuration
        @param delay: Delay between sending each data block
        @return: None
        """
        first_fast_packets = 0
        count = 0

        leftover = len(audio_file) % data_block_size
        update_s_level_delay = 0

        for i in range(len(audio_file) // data_block_size):

            if not self.__exit_flag:
                data_block = audio_file[i * data_block_size: (i + 1) * data_block_size]
                # self.__address contains both the address and the port
                self.__audio_send_socket.sendto(data_block, (self.__client_ip, self.__send_port))

                if self.__input_sound_level_updater_thread is not None:
                    if update_s_level_delay % 50 == 0:
                        sound_level = audioop.rms(data_block, 2)
                        self.__input_sound_level_updater.set_new_sound_level(direction="input",
                                                                             new_value=sound_level)
                    update_s_level_delay += 1

                if first_fast_packets < burst_pkgs:
                    time.sleep(0.002)
                    first_fast_packets += 1
                else:
                    time.sleep(delay)  # data_block_size//2//8 ms

                count = count + data_block_size
                # time.sleep(delay)  # data_block_size//2//8 ms
            else:
                logger.warning("File playing stopped prematurely.")
                self.__paging_state = False
                break

        if leftover != 0 and not self.__exit_flag:
            data_block = audio_file[len(audio_file) - leftover:]
            self.__audio_send_socket.sendto(data_block, (self.__client_ip, self.__send_port))
            count = count + leftover

        logger.info(f"Sent this many bytes: {count / 2}, out of {len(audio_file) / 2}")

    def _reply_with_away_message(self, audio_file_path):
        self.__running_state_flag = True

        logger.info(f"Start sending away message unicast to {self.__client_ip} .")
        # logger.info(f"File to reply with: {audio_file_path}")

        try:
            fp = open(audio_file_path, 'rb')
            audio_file = fp.read()
            fp.close()
            # convert the data to 16-bit signed
            # audio_file_pcm16 = audioop.ulaw2lin(audio_file, 2)
        except Exception as ex:
            warning = f"Unable to read the given {audio_file_path} into bytes: {ex}"
            logger.warning(warning)
            self.__running_state_flag = False
            raise Exception(warning)

        data_block_size, delay, burst_pkgs = self.__configure_sending_params()

        # leftover = len(audio_file) % data_block_size
        # count = 0

        if self.__audio_format == "ulaw" or self.__audio_format == "alaw":

            logger.info("Using %s audio format at %s Hz", self.__audio_format, self.__frequency)
            logger.info(f"Sending {data_block_size / 2} byte packages with {delay}s delay")

            self.__transmit_data(audio_file=audio_file,
                                 burst_pkgs=burst_pkgs,
                                 data_block_size=data_block_size,
                                 delay=delay)

        else:
            logger.error(f"Audio format [{self.__audio_format}] not recognised!!!")
            raise Exception("Unsupported audio format")

        # Signal the paging master when finished with playing
        self.__running_state_flag = False

    def _play_audio(self):
        """
            Used to play the audio data received from the client
        @return:
        """
        self.__running_state_flag = True
        logger.info(f"Start receiving audio from client at {self.__client_ip}")
        audio_output_dev = self._get_audio_device(is_input=False)
        # self._init_message_queue(audio_output_dev)

        self.is_connection_lost = False

        with self.__audio_recv_socket:
            self.__audio_recv_socket.bind(("0.0.0.0", self.__recv_port))

            self.__audio_recv_socket.settimeout(1)
            # logger.warning("Receive socket: %s", self.__audio_recv_socket)
            # logger.warning("Receive socket port: %s", self.__recv_port)

            while not self.__exit_flag:
                num_listen_socket_timeouts = 0
                self._listen_message_q_buff_msg = 0
                while self.listen:
                    # This way it is initialised every time we get in Listen mode
                    if self._listen_message_q is None:
                        self._init_message_queue(audio_output_dev)
                    try:
                        data = self.__audio_recv_socket.recv(1600)
                        if data:
                            self._listen_message_q_buff_msg += 1
                            self._listen_message_q.put(data)  # Add data in the queue

                            if self._listen_message_q_buff_msg == 3:
                                self._listen_message_q_thread.start()
                            time.sleep(0.001)

                    except socket.timeout:
                        num_listen_socket_timeouts += 1
                        if num_listen_socket_timeouts > 5:
                            logger.warning(f"Listen socket timed-out {num_listen_socket_timeouts} times")
                            self.is_connection_lost = True
                # Used in Half duplex mode
                if self.talk and self._listen_message_q is not None:
                    time.sleep(0.02)
                    self._close_message_queue(audio_output_dev)
                time.sleep(0.001)  # NOTE: Prevents excessive CPU usage in case listen is False.

        self._close_message_queue(audio_output_dev)
        self.__running_state_flag = False
        self.__audio_recv_socket.close()
        audio_output_dev.close()

    def _play_from_queue(self, output_audio_device):
        self.packages_read = 0

        update_s_level_delay = 0
        # while self._listen_message_q.empty() is False:
        while not self.__exit_flag:
            if self.listen:
                try:
                    audio_bin_buff = self._listen_message_q.get()  # Get next in line data
                    self.packages_read += 1
                    output_audio_device.write(audio_bin_buff)  # Write the data to the alsa output device

                    if self.__output_sound_level_updater_thread is not None:
                        if update_s_level_delay % 50 == 0:
                            sound_level = audioop.rms(audio_bin_buff, 2)
                            self.__output_sound_level_updater.set_new_sound_level(direction="output",
                                                                                  new_value=sound_level)
                        update_s_level_delay += 1
                    self._listen_message_q.task_done()

                except Exception as e:
                    # self._listen_message_q.task_done()
                    logger.warning(f"Following error occurred: {e}")
                    break
            else:
                break

            # time.sleep(1)
            time.sleep(0.001)

        logger.debug("Finished reading from the queue. Packages played: %s", self.packages_read)
        self.__output_sound_level_updater.set_new_sound_level(direction="output", new_value=0)

    def _init_message_queue(self, audio_device):
        if self._listen_message_q is None:
            self._listen_message_q = Queue(self.__Q_SIZE)
            # Thread to be started only after there has been data written into the queue
            self._listen_message_q_thread = Thread(
                target=self._play_from_queue,
                kwargs={"output_audio_device": audio_device},
                daemon=False,
            )
            self._listen_message_q_thread.setName("Thread-IncomingAudioQueue")
            # logger.info("Message que initialised")

    def _close_message_queue(self, audio_output_dev):

        logger.debug("Packages received: %s", self._listen_message_q_buff_msg)
        while self.packages_read != self._listen_message_q_buff_msg:
            # Play the final packages
            logger.warning("Playing final packages")
            audio_bin_buff = self._listen_message_q.get()  # Get next in line data
            audio_output_dev.write(audio_bin_buff)  # Write the data to the alsa output device
            self.packages_read += 1
            self._listen_message_q.task_done()

        self._listen_message_q_buff_msg = 0
        # Close the message queue
        if self._listen_message_q is not None:
            self._listen_message_q.should_abort_immediately = True
            self._listen_message_q.join()
            # logger.info("--->>> Queue joined() <<<---")
            self._listen_message_q = None
        else:
            logger.info("Message queue is already None")

        # Close the reading thread
        if self._listen_message_q_thread is not None:
            # TODO: Find out why the thread cannot be joined every time
            self._listen_message_q_thread.join(3)
            # logger.warning("Thread alive? -> %s", self._listen_message_q_thread.is_alive())
            # logger.info("--->>> Thread joined() <<<---")
            self._listen_message_q_thread = None
        else:
            logger.info("Message queue thread is already None")
