import logging
import logging.config
import time
from threading import Thread

import alsaaudio
import audioop
from master.utils import audio_uci_to_config_dict
from utils import get_logger
from server.services.sound_level_updater import SoundLevelUpdater, get_sound_level

logger = get_logger("PageAudioService")


class PageAudioService:
    def __init__(self):
        # configure logger
        # logging.config.fileConfig("logging.conf")
        # self.logger = logging.getLogger("PageAudioService")
        try:
            config = audio_uci_to_config_dict()
        except Exception as e:
            logger.error("PageAudioService initialization failed!")
            raise Exception("PageAudioService initialization failed!")

        self.__audio_format = config["audio_format"]
        if self.__audio_format == "alaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_A_LAW
        elif self.__audio_format == "ulaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_MU_LAW
        else:
            raise Exception(
                "Audio codec not recognized by Page audio service,  valid option 'ulaw' and 'alaw' "
            )

        self.__input_sound_level_updater = SoundLevelUpdater()

        self.__exit_flag = False
        self.__record_device = config["input_device"]
        self.__frequency = config["sample_rate"]
        self.__alsa_period_size = 256
        self.__num_channel = config["num_channels"]
        self.__file_to_play = None
        self.__address = None  # Contains both the address and the audio port
        self.__audio_socket = None
        self.is_alsa_capture_lost = False
        self.__paging_state = False

        # Thread handles
        self.__audio_sending_thread = None
        self.__input_sound_level_updater_thread = None

    @property
    def paging_state(self):
        return self.__paging_state

    @property
    def get_file_to_play(self):
        return self.__file_to_play

    def set_file_to_play(self, audio_file_path):
        self.__file_to_play = audio_file_path
        logger.info(f"Configured file to play: {self.__file_to_play}")

    # Address contains both address and port
    def update_config(self, audio_socket, address):
        self.__address = address
        self.__audio_socket = audio_socket
        self.__exit_flag = False
        self.is_alsa_capture_lost = False

    def _get_audio_device(self):
        try:
            audio_input = alsaaudio.PCM(
                type=alsaaudio.PCM_CAPTURE,
                mode=alsaaudio.PCM_NORMAL,  # PCM_NONBLOCK
                device=self.__record_device,
            )
            # Keep this value relatively low when using PCM_NORMAL. Irrelevant with PCM_NONBLOCK
            audio_input.setperiodsize(self.__alsa_period_size)
            audio_input.setrate(self.__frequency)
            audio_input.setchannels(self.__num_channel)
            audio_input.setformat(self.__audio_codec)
            return audio_input
        except Exception as e:
            logger.error(f"Failure: Audio device inaccessible: {e}")
            raise Exception("PageAudioService Failed!")

    def _record_and_send_audio(self):
        """
            Used for the live paging functionality
        @return:
        """
        self.__paging_state = True
        logger.info("Start audio capture and multicast/broadcast.")
        audio_input = self._get_audio_device()

        num_no_capture = 0
        prev_capture = True
        self.is_alsa_capture_lost = False

        logger.debug("Sending data to this address: {}".format(self.__address))

        update_s_level_delay = 0

        while not self.__exit_flag:
            l, data = audio_input.read()
            if l:
                self.__audio_socket.sendto(data, self.__address)
                if self.__input_sound_level_updater_thread is not None:
                    if update_s_level_delay % 200 == 0:
                        # logger.warning("Data length: %s", l)
                        sound_level = audioop.rms(data, 2)
                        self.__input_sound_level_updater.set_new_sound_level(direction="input",
                                                                             new_value=sound_level)
                    update_s_level_delay += 1
                prev_capture = True
            else:
                num_no_capture = 0 if prev_capture else num_no_capture + 1
                prev_capture = False
            if num_no_capture > 1000:
                logger.warning(
                    f"No data received from alsa capture, for over {num_no_capture} frames"
                )
                self.is_alsa_capture_lost = True
                num_no_capture = 0
            time.sleep(0.001)

        audio_input.close()
        audio_input = None
        self.__paging_state = False
        logger.info("Stop audio capture and multicast/broadcast.")

    def __configure_sending_params(self):
        """
            Calculates the data block size and delay between sending each block based on the device configuration.
        @return: data_block_size, delay
        """
        burst_pkgs = 5
        if self.__frequency == 8000:
            data_block_size = 960
            delay = 0.12
            # delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000)) - 0.001
        elif self.__frequency == 24000:

            data_block_size = 964  # 1280
            delay = float("{0:.4f}".format(data_block_size / 2 / self.__frequency))
            burst_pkgs = 3
        else:
            data_block_size = 1280
            delay = 0.079
            logger.warning("Sample rate not supported. Using default values")
        return data_block_size, delay, burst_pkgs

    def __transmit_data(self, audio_file: bytes, burst_pkgs: int, data_block_size: int, delay: float):
        """
            This function sends the decoded audio data over to the client
        @param audio_file: Decoded audio data in bytes
        @param burst_pkgs: Number of fast packages at the beginning of transmission
        @param data_block_size: Data block size, calculated based on the device configuration
        @param delay: Delay between sending each data block
        @return: None
        """

        first_fast_packets = 0
        count = 0

        leftover = len(audio_file) % data_block_size

        for i in range(len(audio_file) // data_block_size):

            if not self.__exit_flag:
                data_block = audio_file[i * data_block_size: (i + 1) * data_block_size]
                self.__audio_socket.sendto(data_block, self.__address)

                if first_fast_packets < burst_pkgs:
                    time.sleep(0.002)
                    first_fast_packets += 1
                else:
                    time.sleep(delay)  # data_block_size//2//8 ms

                count = count + data_block_size
                # time.sleep(delay)  # data_block_size//2//8 ms
            else:
                logger.warning("File playing stopped prematurely.")
                self.__paging_state = False
                break

        if leftover != 0 and not self.__exit_flag:
            data_block = audio_file[len(audio_file) - leftover:]
            self.__audio_socket.sendto(data_block, self.__address)
            count = count + leftover

        logger.info(f"Sent this many bytes: {count / 2}, out of {len(audio_file) / 2}")

    def _send_recorded_audio_message(self):
        logger.info("Start paging from recorded file (multicast/broadcast).")
        logger.info(f"File to start paging: {self.__file_to_play}")
        audio_file = None
        self.__paging_state = True
        self.is_alsa_capture_lost = False  # No alsa capture is used here

        try:
            fp = open(self.__file_to_play, 'rb')
            audio_file = fp.read()
            fp.close()
            # convert the data to 16-bit signed
            audio_file_pcm16 = audioop.ulaw2lin(audio_file, 2)
        except Exception as ex:
            logger.warning("Unable to read the given %s into bytes: %s", self.__file_to_play, ex)

        logger.debug("Sending data to this address: {}".format(self.__address))

        data_block_size, delay, burst_pkgs = self.__configure_sending_params()

        if self.__audio_format == "ulaw" or self.__audio_format == "alaw":

            # logger.info("Using %s audio format at %s Hz", self.__audio_format, self.__frequency)
            # logger.info(f"Sending {data_block_size / 2} byte packages with {delay}s delay")

            self.__transmit_data(audio_file, burst_pkgs, data_block_size, delay)

        else:
            logger.error(f"Audio format [{self.__audio_format}] not recognised!!!")
            raise Exception("Unrecognised audio codec")

        # Signal the paging master when finished with playing
        self.__file_to_play = None
        self.__paging_state = False

    def exit(self):
        self.__exit_flag = True
        if self.__audio_sending_thread is not None:
            self.__audio_sending_thread.join()
            self.__audio_sending_thread = None
        if self.__input_sound_level_updater_thread is not None:
            self.__input_sound_level_updater.exit()
            self.__input_sound_level_updater_thread.join()
            self.__input_sound_level_updater_thread = None
            # logger.info("Closed __input_sound_level_updater_thread")

        self.__file_to_play = None
        self.__paging_state = False
        logger.info("Exit PagingAudioService")
        self.__exit_flag = False

    def run(self):
        # Note: This function is idempotent.
        if self.__audio_sending_thread is None:
            logger.info("Start PagingAudioService")
            if self.__file_to_play is None:
                self.__audio_sending_thread = Thread(target=self._record_and_send_audio)
            else:
                self.__audio_sending_thread = Thread(target=self._send_recorded_audio_message)
            self.__audio_sending_thread.daemon = True
            self.__audio_sending_thread.start()
        else:
            logger.warning(
                "Attempt to start PagingAudioService. Already running. Ignored"
            )

        if self.__input_sound_level_updater_thread is None:
            self.__input_sound_level_updater_thread = Thread(target=self.__input_sound_level_updater.run)
            self.__input_sound_level_updater_thread.daemon = True
            self.__input_sound_level_updater_thread.setName("Thread-InputSoundLvlUpdater")
            self.__input_sound_level_updater_thread.start()
