import socket
from datetime import datetime

# from src.utils import get_uci
from utils import get_uci, get_logger
from shared_constants import PAGING_MASTER_TMP_PATH, IF_CONFIG_FILE

logger = get_logger("Master/Utils")


def get_active_devices(barp, listen_status_socket, pmaster, active_devices):
    while not pmaster.exit_flag:
        try:
            data, address = listen_status_socket.recvfrom(4096)
        except socket.timeout:
            continue
        msg = barp.decode(data)
        if msg[0]["message_type"] == 128:
            active_devices[msg[0]["sender_id"]] = {
                "state": msg[1]["state"],
                "time_stamp": datetime.now(),
            }

        # Removal of devices with stale status.
        current_time = datetime.now()
        for dev_id in list(active_devices.keys()):
            if (current_time - active_devices[dev_id]["time_stamp"]).seconds > 5:
                del active_devices[dev_id]
    print("Status listening stopped!")


def is_multicast(ip):
    """
    Performs check on ipv4 ip address to determine if it is multicast or not.
    Note: it is assumed that the address is valid.
    Args:
        ip (str)
    Returns:
        (bool) : True if address is multicast and false if it isn't.
    """
    ip_address = [int(i) for i in ip.split(".")]
    if 224 <= ip_address[0] <= 239:
        return True
    else:
        return False


def is_broadcast(ip, broadcast_ip):
    """
    Performs check on ipv4 ip address to determine if it is multicast or not.
    Note: it is assumed that the address is valid.
    Args:
        ip (str)
        broadcast_ip (str)
    Returns:
        (bool) : True if address is multicast and false if it isn't.
    """
    # logger.info("Checking IP: {}".format(ip))

    if ip == "0.0.0.0":
        return True
    elif ip == "255.255.255.255":
        return True
    elif ip == broadcast_ip:
        return True
    else:
        return False


def load_ip_config() -> dict:
    """
        Creates a file inside the RAM of the device, containing its IP settings.
    Returns:
        dict with the ip settings
    """
    # Only used in this function, so no need to have it available for the whole file
    import re
    from subprocess import Popen, PIPE, TimeoutExpired
    from pathlib import Path
    import os

    ip_settings = dict()

    # declaring the regex pattern for IP addresses
    pattern = re.compile('''((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)''')

    # Make sure the PAGING_MASTER_TMP_PATH exists
    if not os.path.exists(PAGING_MASTER_TMP_PATH):
        os.makedirs(PAGING_MASTER_TMP_PATH)

    # Second get the file or create it if it doesn't exist
    file_with_path = Path(PAGING_MASTER_TMP_PATH + IF_CONFIG_FILE)
    file_with_path.touch(exist_ok=True)

    settings_file = open(file_with_path, "w")
    command = ["ifconfig", "eth0"]
    process = Popen(command, stdout=settings_file, stderr=PIPE, )
    try:
        outs, errs = process.communicate(timeout=5)
        settings_file.close()  # Close the fd after the data has been added to it

        # logger.info(f"Sent command: {command}")
        # logger.info(f"Result from the command: {outs}")

        # Second open the file and extract the ip, netmask and broadcast
        with open(file_with_path, 'r') as f:
            file = f.readlines()
            valid = []
            # Find lines with IPs
            for line in file:
                line = line.rstrip()
                result = pattern.search(line)

                # valid IP addresses
                if result:
                    valid.append(line)
            valid = valid[0].strip().replace("  ", " ").split(" ")  # Clean up the data from unnecessary spaces

            # Store the values to key-value pairs
            for index in range(len(valid)):
                if index % 2 == 0:
                    ip_settings[valid[index]] = valid[index+1]

            # logger.info(f"Valid IP addresses: {valid}")

    except TimeoutExpired:
        process.kill()
        outs, errs = process.communicate()
        print(f"Timed out when loading IP config. Errs :{errs}")
    except Exception as e:
        logger.info(f"Following error occurred: %s", e)
    # logger.info(f"Returning following settings: {ip_settings}")
    return ip_settings


def master_uci_to_config_dict():

    config_dict = {}
    package = "icpaging-master"
    config_dict["id"] = int(get_uci(package, "barp", "id"))
    config_dict["ip"] = get_uci(package, "barp", "ip_addr")
    config_dict["ports"] = {
        "control": int(get_uci(package, "barp", "control_port")),
        "audio": int(get_uci(package, "barp", "audio_port")),
        "status": int(get_uci(package, "barp", "status_port")),
        "master_audio": int(get_uci(package, "barp", "master_audio")),
    }
    config_dict["encoding_type"] = int(get_uci(package, "barp", "encoding"))
    config_dict["multicast_group_address"] = get_uci(package, "barp", "mcast_grp_addr")
    config_dict["relay_activation_time"] = int(
        get_uci(package, "barp", "relay_activation_time")
    )
    config_dict["barp_header_kwargs"] = {
        "flags": int(get_uci(package, "barp_args", "flags")),
        "sequence_number": int(get_uci(package, "barp_args", "sequence_number")),
    }
    # Returns this kind of dict: {'inet': '192.168.10.65', 'netmask': '255.255.255.0', 'broadcast': '192.168.10.255'}
    config_dict["device_ip_settings"] = load_ip_config()
    return config_dict


def audio_uci_to_config_dict():
    """Creates config from uci to be used by Audio service of PagingMaster"""
    config_dict = {}
    package = "icpaging-master"
    config_dict["input_device"] = get_uci(package, "audio", "input_device")  # plug:dsnoop_analog was not working
    config_dict["output_device"] = get_uci(package, "audio", "output_device")  # plug:master1 was not working
    config_dict["sample_rate"] = int(get_uci(package, "audio", "sample_rate"))
    config_dict["audio_format"] = get_uci(package, "audio", "audio_format")
    config_dict["period_size"] = int(get_uci(package, "audio", "period_size"))
    config_dict["num_channels"] = int(get_uci(package, "audio", "num_channels"))
    return config_dict

