# import logging
# import logging.config
from nextion.nextion_enum import KeyID, NextionPagingState
from nextion.mapping import STATION_MAPPING, GROUP_MAPPING
from nextion.state_variables import ButtonState, NextionStateVariable
import threading
from datetime import datetime
from utils import get_logger

logger = get_logger("NextionInterface")


# Ensures that there is only one instance of the interface
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NextionInterface(metaclass=Singleton):
    MESSAGE_LIFE_TIME = 2  # used by Nextion interface to make message stale

    def __init__(
        self,
        nextion_state: NextionStateVariable,
        buttons_state: ButtonState,
        half_duplex: bool = False,
    ):
        # configure logger
        # logging.config.fileConfig("../logging.conf")
        # self.logger = logging.getLogger("NextionInterface")

        self.nextion_state = nextion_state
        self.buttons_state = buttons_state
        self.__half_duplex = half_duplex
        self.__call_info_changed = False
        self.__half_duplex_listen = True
        self.__call_info = {}
        self.__message = ""
        self.__message_changed = True
        self.__message_time = datetime.now()  # Time at which message was last set.
        self.__relay = False

        # Used to notify the state_change_recorder to get the state
        self.__state_was_changed = False

        self.lock = threading.Lock()

    @property
    def message(self):
        self.__message_changed = False
        return self.__message

    @message.setter
    def message(self, msg):
        self.__message_time = datetime.now()
        self.__message_changed = True
        self.__message = msg

    @property
    def message_changed(self):
        """Flag denoting if the message s changed or not.

        Returns:
            bool: True if the message is changed
            and not yet accessed through message property
        """
        message_is_not_stale = (
            datetime.now() - self.__message_time
        ).seconds < NextionInterface.MESSAGE_LIFE_TIME
        message_is_valid = self.__message != ""
        if message_is_valid:

            if message_is_not_stale or self.nextion_state.state in [
                NextionPagingState.IN_ACTIVE_CALL,
                NextionPagingState.INCOMING_CALL_REQUEST,
                NextionPagingState.MESSAGE,
                NextionPagingState.PAGING_ACTIVE,
            ]:
                # These two are the only exceptions.
                return self.__message_changed

            elif message_is_not_stale or self.nextion_state.state in [
                NextionPagingState.UNATTENDED_MODE,
                NextionPagingState.CALL_FORWARDING_MODE
            ]:
                # logger.info("The screen should go off, but the text should stay displayed")
                return False

            else:
                # This case will only happen after the message went stale
                # example: call rejection, failed call etc.
                self.__message = ""
                return True
        else:
            return self.__message_changed

    @property
    def relay(self):
        """Calling it resets it. To be only called by paging server."""
        if self.__relay:
            self.__relay = False
            return True
        else:
            return False

    @relay.setter
    def relay(self, value):
        """To be set by nextion panel only.

        Args:
            value ([bool])
        """
        self.__relay = value

    @property
    def is_half_duplex(self):
        return self.__half_duplex

    @property
    def call_info(self):
        return self.__call_info["id"], self.__call_info["ip"]

    @property
    def is_half_duplex_listen_mode(self):
        return self.__half_duplex_listen

    @property
    def call_info_changed(self):
        return self.__call_info_changed

    @property
    def selected_clients(self):
        clients = []
        try:
            for key in KeyID:
                if "STATION" in key.name and self.buttons_state.get_state(key):
                    clients.append(STATION_MAPPING[key.name])
                if "GROUP" in key.name and self.buttons_state.get_state(key):
                    clients += GROUP_MAPPING[key.name]
            # TODO: Map the station and groups to actual client ids.
            # logger.warning(f"Configure groups to client mapping!")
        except Exception as e:
            logger.warning(f"Following exception occurred: {e}")
        return sorted(list(set(clients)))

    def set_half_duplex_mode(self, listen=True):
        self.__call_info_changed = True
        logger.info(f"Listen mode is {listen}")
        self.__half_duplex_listen = listen

    def set_half_duplex(self):
        self.__half_duplex = True

    def set_full_duplex(self):
        self.__half_duplex = False

    def update_call_info(self, caller_id, caller_ip):
        self.__call_info["id"] = caller_id
        self.__call_info["ip"] = caller_ip

    @property
    def state_change(self):
        return self.__state_was_changed

    def set_state_change(self, value: bool = True):
        self.__state_was_changed = value

    def clear_state_change(self):
        self.__state_was_changed = False
