# import logging
# import logging.config
from datetime import datetime
from threading import Thread
# from matplotlib.pyplot import get

import serial

from nextion.nextion_enum import KeyID, KeyStatus, NextionPagingState
from nextion.nextion_interface import NextionInterface
from nextion.state_variables import ButtonState, NextionStateVariable
from utils import get_logger, mcu_gpio_control

logger = get_logger("NextionPanel")
serial_logger = get_logger(name="NextionPanel-Serial", log_file='/var/log/serial_message_logger')


# Ensures that there is only one instance of the panel
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NextionPanel(metaclass=Singleton):

    TERMINATING_CHARACTER = b"\xff\xff\xff"

    def __init__(
        self,
        nextion_interface: NextionInterface,
        baud_rate: int,
        serial_port: str,
        timeout: int = 0.5,
        number_pages: int = 0,
    ):
        self.__number_pages = number_pages
        self.__baud_rate: int = baud_rate
        self.__serial_port: str = serial_port
        self.__timeout: int = timeout
        self.__read_flag: bool = False
        self.__process_is_running = False
        # keeps track of the time last awake message was sent.
        self.__last_awake = datetime.now()

        # These variables are used outside the class in paging_server through nextion_interface.
        # Note: Handle them with care.
        self.buttons_state: ButtonState = nextion_interface.buttons_state
        self.nextion_state: NextionStateVariable = nextion_interface.nextion_state
        self.nextion_interface: NextionInterface = nextion_interface
        self.restart_main_app_flag = False

    @property
    def baud_rate(self):
        return self.__baud_rate

    @property
    def serial_port(self):
        return self.__serial_port

    @property
    def timeout(self):
        return self.__timeout

    @property
    def active_stations_and_groups(self):
        """
            Returns active groups and stations as a tuple of two lists (stations, groups)
        """
        stations = []
        groups = []
        for key in KeyID:
            if "STATION" in key.name and self.buttons_state.get_state(key):
                stations.append(key.value)
            if "GROUP" in key.name and self.buttons_state.get_state(key):
                groups.append(key.value)
        return stations, groups

    def run(self):
        if not self.__process_is_running:
            self.__process_is_running = True
            # Turn off paging button LED on start
            mcu_gpio_control("gpio14", False)

            self.read_data()
        else:
            raise Exception("Process already running.")

    def stop(self):
        if self.__process_is_running:
            self.read_data(stop=True)

        # Turn on paging button LED on exit
        mcu_gpio_control("gpio14", True)

    def read_data(self, stop: bool = False):
        if self.__read_flag and not stop:
            logger.warning("Already reading the data from nextion serial port.")
        elif stop:
            self.__read_flag = not stop
        else:
            self.__read_flag = not stop
            with serial.Serial(
                self.serial_port, self.baud_rate, timeout=self.timeout
            ) as ser:
                self.sync_leds(ser)
                self.sync_state(ser)
                self.configure_call_mode(ser)

                while self.__read_flag:
                    data = ser.read_until(
                        terminator=NextionPanel.TERMINATING_CHARACTER, size=None
                    )
                    if len(data):
                        self.process_data(data)
                        self.sync_leds(ser)
                    self.sync_state(ser)
            logger.info("Reading data from serial interface finished!")

    def configure_call_mode(self, ser):
        message = (
            f"hdx.val={int(self.nextion_interface.is_half_duplex)}".encode()
            + NextionPanel.TERMINATING_CHARACTER
        )
        # logger.info(f"Configured call mode {message}")
        ser.write(message)

    def process_data(self, data):
        """
        Processes the incoming data from nextion panel and updated the states of the buttons.
        Currently only commands starting with # are supported. Rest all are ignored.

        Args:
            data (bytes): Data from panels via serial port. The data looks as follows.
                            data[0]     --> "#" - command start marker (0x06)
                            data[1]     --> Command length (L) in bytes
                            data[2]     --> Command Type ("K"-keys)'
                            data[3:L-1] --> command data (specific for the command)
                            data[L:L+2] --> Command   termination

        Raises:
            Exception: If a command type other than 'K' is encountered.

        Returns:
            data Optional[]
        """
        try:
            serial_logger.info(f"Data read from the panel: {data}")
            if data[0] == ord("#"):
                length = data[1]
                command_type = data[2]
                if command_type == ord("K"):
                    key_status = {"id": KeyID(data[4]), "status": KeyStatus(data[3])}
                    logger.debug(key_status)
                    self.process_key_status(key_status)
                    logger.debug(self.nextion_state)
                else:
                    raise Exception("Unrecognized command type!")
            elif data[0] == ord("\x68"):
                logger.info("Touch event 0x68 (awake)")
            elif data[0] == ord("\x67"):
                logger.info("Touch event 0x67 (sleep)")
            elif data[0] == ord("\x70"):
                logger.debug(f"Receive requested text data (0x70):{data}")
            elif data[0] == ord("\x71"):
                logger.debug(f"Receive requested text data (0x71):{data}")
            elif data[0] == ord("\x86"):
                logger.debug("Nextion Panel entering in sleep mode")
            elif data[0] == ord("\x87"):
                logger.debug("Nextion Panel waking up from sleep")
            elif data[0] == ord("\x1A"):
                pass
                # logger.warning("Invalid var/attr!")
            elif data[0] == ord("\x1C"):
                logger.debug("Wrong assignment!")
            elif data[0] == ord("\x88"):
                logger.warning(f" Nextion panel reset [{data}], restarting the app. Please wait...: ")
                self.restart_main_app_flag = True
            else:
                logger.warning(f"Unrecognized command : {data}")
        except Exception as e:
            logger.warning("Failed to read incoming data from nextion panel serial port")
            logger.warning(e)

            import traceback
            logger.warning(traceback.format_exc())

    def process_key_status(self, key_status):
        # logger.info(f"Inside the process_key_status(). {key_status}")
        status, key_id = key_status["status"], key_status["id"]

        if key_id == KeyID.CANCEL_SELECTION_BUTTON:
            self.process_cancel_selection_button(key_id, status)
        elif key_id == KeyID.PAGE_BUTTON:
            self.process_page_button(key_id, status)
        elif "STATION" in key_id.name or "GROUP" in key_id.name:
            self.process_station_group_button(key_id, status)
        elif key_id == KeyID.RECORD_BUTTON:
            self.process_record_button(key_id, status)
        elif key_id == KeyID.AWAY_BUTTON:
            self.process_away_button(key_id, status)
        elif key_id == KeyID.RELAY_BUTTON:
            self.process_relay_button(key_id, status)
        elif key_id == KeyID.ACCEPT_CALL_BUTTON:
            self.process_accept_call_button(key_id, status)
        elif key_id == KeyID.CLOSE_CALL_BUTTON:
            self.process_close_call_button(key_id, status)
        elif key_id == KeyID.AWAY_MESSAGE_CHECKBOX:
            self.process_away_message_checkbox(key_id, status)
        elif key_id == KeyID.CUSTOM_MESSAGE_CHECKBOX:
            self.process_custom_message_checkbox(key_id, status)
        else:
            # TODO: Incorporate state and other buttons
            pass
            raise NotImplementedError

    def awake_nextion_panel(self, ser):
        if (datetime.now() - self.__last_awake).seconds > 5:
            self.__last_awake = datetime.now()
            awake_message = (
                f"hdx.val={self.nextion_interface.is_half_duplex*1}".encode()
                + NextionPanel.TERMINATING_CHARACTER
            )
            logger.info(f"Waking up the display")
            ser.write(awake_message)

    def sync_state(self, ser: serial.Serial):
        """Syncs the state of the app to Nextion panel
        Args:
            ser (serial.Serial): [description]
        """
        message = ""
        with self.nextion_interface.lock:

            if self.nextion_state.state_changed:
                message = f'fwstate.val={self.nextion_state.state.value}'
                ser.write(message.encode() + NextionPanel.TERMINATING_CHARACTER)
                # Notify the state change recorder, and clear it there
                self.nextion_interface.set_state_change(True)
                self.nextion_state.state_synced()

            if self.nextion_interface.message_changed:
                message = f't0.txt="{self.nextion_interface.message}"'
                logger.info(f"State Sync {message}")
                ser.write(message.encode() + NextionPanel.TERMINATING_CHARACTER)
                self.nextion_state.state_synced()
            if self.nextion_state.state in [
                NextionPagingState.IN_ACTIVE_CALL,
                NextionPagingState.INCOMING_CALL_REQUEST,
                NextionPagingState.PAGING_ACTIVE,
                NextionPagingState.RECORDING_AWAY_MESSAGE,
                NextionPagingState.RECORDING_CUSTOM_MESSAGE
            ]:
                self.awake_nextion_panel(ser)

            if self.nextion_state.state in [
                NextionPagingState.UNATTENDED_MODE,
                NextionPagingState.CALL_FORWARDING_MODE
            ]:
                # logger.info("Screen has to go off while the text stays on the screen")
                self.nextion_state.state_synced()

    def sync_leds(self, ser: serial.Serial):
        """Sends the commands to sync leds on the nextion panel based on the state of buttons.
        Example sends following to the panel if a button is pressed.
        b"leds1=1\xff\xff\xff\xff"  --> This will (select) set button 0 and deselect rest of them.
        b"leds1=2\xff\xff\xff\xff"  --> This will (select) set button 0 and 1 and deselect rest of them.

        Note:
            leds1 control buttons 0 to 15.
            leds2 control buttons 16 to 31.
            leds3 control buttons 32 to 47.
            leds4 contain buttons 48 to 63.
            leds5 contain buttons 64 to 79.
            leds6 contain buttons 80 to 95.
            leds7 contain buttons 96 to 111.

        """
        if self.buttons_state.led1_changed:
            ser.write(self.buttons_state.leds1 + NextionPanel.TERMINATING_CHARACTER)
        if self.buttons_state.led2_changed:
            ser.write(self.buttons_state.leds2 + NextionPanel.TERMINATING_CHARACTER)
        if self.buttons_state.led3_changed:
            ser.write(self.buttons_state.leds3 + NextionPanel.TERMINATING_CHARACTER)
        if self.buttons_state.led4_changed:
            ser.write(self.buttons_state.leds4 + NextionPanel.TERMINATING_CHARACTER)

        if self.__number_pages >= 1:
            if self.buttons_state.led5_changed:
                ser.write(self.buttons_state.leds5 + NextionPanel.TERMINATING_CHARACTER)
            if self.buttons_state.led6_changed:
                ser.write(self.buttons_state.leds6 + NextionPanel.TERMINATING_CHARACTER)
            if self.buttons_state.led7_changed:
                ser.write(self.buttons_state.leds7 + NextionPanel.TERMINATING_CHARACTER)

    def process_away_message_checkbox(self, key_id, status):
        with self.nextion_interface.lock:
            # logger.warning("Current state: %s", self.nextion_state.state)
            if (  # When the away message should be used
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                logger.info("Setting state to USE_AWAY_MESSAGE")
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.USE_AWAY_MESSAGE
            elif (  # When both the Recording and the away message checkbox are pressed
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.MESSAGE
            ):
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.RECORDING_AWAY_MESSAGE
            elif (  # No state transition if the checkbox is pressed while recording
                    status == KeyStatus.PRESSED and
                    self.nextion_state.state == NextionPagingState.RECORDING_AWAY_MESSAGE
            ):
                self.buttons_state.set_button(key_id)
            elif (  # Stop playing the message
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.USE_AWAY_MESSAGE
            ):
                logger.info("Setting state to FINISH_OPERATION")
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.FINISH_OPERATION
            elif (  # No state transition if the checkbox is released while recording
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.RECORDING_AWAY_MESSAGE
            ):
                self.buttons_state.reset_button(key_id)
            elif (  # Uncheck when the recording is done
                    status == KeyStatus.RELEASED and
                    self.nextion_state.p_state == NextionPagingState.FINISH_OPERATION
            ):
                self.buttons_state.reset_button(key_id)
            elif (
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.USE_AWAY_MESSAGE
            ):
                self.buttons_state.reset_button(key_id)
            elif (  # Just reset the checkbox. This state is after a message is stopped prematurely
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                self.buttons_state.reset_button(key_id)
            else:
                self.log_state_action_warning(key_id, status)

    def process_custom_message_checkbox(self, key_id, status):
        with self.nextion_interface.lock:
            logger.warning("Current state: %s", self.nextion_state.state)
            if (  # When the custom message should be sent/locally played
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                logger.info("Setting state to SENDING_CUSTOM_MESSAGE")
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.SENDING_CUSTOM_MESSAGE
            elif (  # When both the Recording and the custom message checkbox are pressed
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.MESSAGE
            ):
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.RECORDING_CUSTOM_MESSAGE
            elif (  # No state transition if the checkbox is pressed while recording
                    status == KeyStatus.PRESSED and
                    self.nextion_state.state == NextionPagingState.RECORDING_CUSTOM_MESSAGE
            ):
                self.buttons_state.set_button(key_id)
            elif (  # Stop sending the message
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.SENDING_CUSTOM_MESSAGE
            ):
                logger.info("Setting state to FINISH_OPERATION")
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.FINISH_OPERATION
            elif (  # No state transition if the checkbox is released while recording
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.RECORDING_CUSTOM_MESSAGE
            ):
                self.buttons_state.set_button(key_id)
            elif (  # Uncheck when the recording is done. State transition initiated from paging server
                    status == KeyStatus.RELEASED and
                    self.nextion_state.p_state == NextionPagingState.FINISH_OPERATION
            ):
                self.buttons_state.reset_button(key_id)
            elif (  # No state transition if the checkbox is released while recording
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.SENDING_CUSTOM_MESSAGE
            ):
                self.buttons_state.reset_button(key_id)
            elif (  # Just reset the checkbox. This state is after a message is stopped prematurely
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                self.buttons_state.reset_button(key_id)
            else:
                self.log_state_action_warning(key_id, status)

    def process_cancel_selection_button(self, key_id, status):
        if (
            status == KeyStatus.PRESSED and
            self.nextion_state.state == NextionPagingState.IDLE
        ):
            self.buttons_state.reset_group_selection()
            self.buttons_state.reset_station_selection()
        else:
            self.log_state_action_warning(key_id, status)

    def process_page_button(self, key_id, status):
        with self.nextion_interface.lock:
            if (
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                # Turn on the paging button LED
                mcu_gpio_control("gpio14", True)
                self.nextion_state.state = NextionPagingState.PAGING_ACTIVE
            elif (
                self.nextion_state.state == NextionPagingState.PAGING_ACTIVE
                and status == KeyStatus.RELEASED
            ):
                # Turn off the paging button LED
                mcu_gpio_control("gpio14", False)
                self.nextion_state.state = NextionPagingState.IDLE
            elif (
                self.nextion_state.state == NextionPagingState.IN_ACTIVE_CALL
                and self.nextion_interface.is_half_duplex
            ):
                if status == KeyStatus.PRESSED:
                    # Turn on the paging button LED
                    mcu_gpio_control("gpio14", True)
                    self.nextion_interface.set_half_duplex_mode(listen=False)
                else:
                    self.nextion_interface.set_half_duplex_mode(listen=True)
                    # Turn off the paging button LED
                    mcu_gpio_control("gpio14", False)
            else:
                self.log_state_action_warning(key_id, status)

    def process_station_group_button(self, key_id, status):
        if self.nextion_state.state == NextionPagingState.IDLE:
            if status == KeyStatus.PRESSED:
                # self.buttons_state.flip_button(key_id.value)  # Used to be this
                self.buttons_state.flip_button(key_id)
        else:
            self.log_state_action_warning(key_id, status)

    def process_record_button(self, key_id, status):
        with self.nextion_interface.lock:
            # logger.warning("Current state: %s", self.nextion_state.state)
            if (
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.IDLE
            ):
                self.buttons_state.set_button(key_id)
                self.nextion_state.state = NextionPagingState.MESSAGE

            elif (  # While recording, keep the buttons pressed
                status == KeyStatus.PRESSED and
                (self.nextion_state.state == NextionPagingState.RECORDING_AWAY_MESSAGE or
                 self.nextion_state.state == NextionPagingState.RECORDING_CUSTOM_MESSAGE)
            ):
                logger.warning("Is the Away checkbox on?")
                self.buttons_state.set_button(KeyID.AWAY_MESSAGE_CHECKBOX)
                self.buttons_state.set_button(key_id)

            elif(  # When the recording needs to stop
                status == KeyStatus.RELEASED and
                (self.nextion_state.state == NextionPagingState.RECORDING_AWAY_MESSAGE or
                    self.nextion_state.state == NextionPagingState.RECORDING_CUSTOM_MESSAGE)
            ):
                self.buttons_state.reset_button(key_id)
                self.nextion_state.state = NextionPagingState.FINISH_OPERATION

            elif (
                status == KeyStatus.RELEASED and
                self.nextion_state.state == NextionPagingState.MESSAGE
            ):
                self.buttons_state.reset_button(key_id)
                self.nextion_state.state = NextionPagingState.IDLE
            else:
                self.log_state_action_warning(key_id, status)

    def process_away_button(self, key_id, status):
        logger.warning("Current state: %s", self.nextion_state.state)

        with self.nextion_interface.lock:
            if (
                status == KeyStatus.PRESSED and
                self.nextion_state.state != NextionPagingState.UNATTENDED_MODE and
                self.nextion_state.state != NextionPagingState.CALL_FORWARDING_MODE
            ):
                self.buttons_state.set_button(key_id)
                self.buttons_state.set_button_status(key_id, status)

                # Turn on the paging button LED
                mcu_gpio_control("gpio14", True)

                self.nextion_state.state = NextionPagingState.UNATTENDED_MODE

            elif (  # Use this to signal the paging server the button was released
                    status == KeyStatus.RELEASED and
                    self.nextion_state.state == NextionPagingState.UNATTENDED_MODE
            ):
                self.buttons_state.set_button_status(key_id, status)
            elif (
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.UNATTENDED_MODE
            ):
                self.buttons_state.reset_button(key_id)

                # Turn off the paging button LED
                mcu_gpio_control("gpio14", False)
                self.nextion_state.state = NextionPagingState.IDLE

            elif (
                status == KeyStatus.PRESSED and
                self.nextion_state.state == NextionPagingState.CALL_FORWARDING_MODE
            ):
                self.buttons_state.reset_button(key_id)
                # Turn off the paging button LED
                mcu_gpio_control("gpio14", False)
                self.nextion_state.state = NextionPagingState.IDLE

            else:
                message = "Case RELEASE and IDLE is irrelevant for process_away_button()"
                self.log_state_action_warning(key_id, status, message)

    def process_relay_button(self, key_id, status):
        with self.nextion_interface.lock:
            if self.nextion_state.state == NextionPagingState.IN_ACTIVE_CALL:
                if status == KeyStatus.PRESSED:
                    self.nextion_interface.relay = True
                    logger.info("Relay activated")
            else:
                self.log_state_action_warning(key_id, status)

    def process_accept_call_button(self, key_id, status):
        # Note: Messages for accepting the call is triggered by PagingServer once the call is started.
        with self.nextion_interface.lock:
            if self.nextion_state.state == NextionPagingState.INCOMING_CALL_REQUEST:
                if status == KeyStatus.PRESSED:
                    self.nextion_state.state = NextionPagingState.IN_ACTIVE_CALL
                    if self.nextion_interface.is_half_duplex:
                        self.nextion_interface.set_half_duplex_mode(listen=True)
            elif self.nextion_state.state == NextionPagingState.IDLE:
                self.nextion_interface.message = "No Incoming call!"
            else:
                self.log_state_action_warning(key_id, status)

    def process_close_call_button(self, key_id, status):
        # Note: Messages for closing call is triggered by PagingServer once the call actually closes.
        with self.nextion_interface.lock:
            if self.nextion_state.state in (
                NextionPagingState.IN_ACTIVE_CALL,
                NextionPagingState.INCOMING_CALL_REQUEST,
            ):
                if status == KeyStatus.PRESSED:
                    # Check if the accept and close buttons are sticky or not!
                    self.nextion_state.state = NextionPagingState.IDLE
                    self.buttons_state.reset_button(KeyID.ACCEPT_CALL_BUTTON)
                    self.buttons_state.set_button(KeyID.CLOSE_CALL_BUTTON)
                # Released does nothing!
                elif status == KeyStatus.RELEASED:
                    self.buttons_state.reset_button(KeyID.CLOSE_CALL_BUTTON)
            # Released does nothing!
            elif status == KeyStatus.RELEASED:
                self.buttons_state.reset_button(KeyID.CLOSE_CALL_BUTTON)
            else:
                self.log_state_action_warning(key_id, status)

    def log_state_action_warning(self, key_id, key_status, particular_message=""):
        if particular_message:
            logger.info(particular_message)
        else:
            logger.warning(
                f"Unrecognized SA Pair: {key_id} is {KeyStatus(key_status)}, {self.nextion_state.state}"
            )
