from time import sleep
import json
import os
from pathlib import Path

from nextion.nextion_interface import NextionInterface
from nextion.nextion_enum import NextionPagingState, KeyID, KeyStatus

from utils import read_json, get_logger
from shared_constants import PAGING_MASTER_TMP_PATH, STATE_FILE

logger = get_logger(name="StateChangeRecorder", log_file="/var/log/nextion_state_changes")


class StateChangeRecorder:

    def __init__(self, nextion_interface: NextionInterface):

        self.nextion_interface = nextion_interface
        self.nextion_state = nextion_interface.nextion_state
        try:
            self.button_state = nextion_interface.buttons_state
        except Exception as e:
            logger.warning(f"Error: {e}")

        # Make sure the PAGING_MASTER_TMP_PATH exists
        if not os.path.exists(PAGING_MASTER_TMP_PATH):
            os.makedirs(PAGING_MASTER_TMP_PATH)

        self._state_file = Path(PAGING_MASTER_TMP_PATH + STATE_FILE)
        self._state_file.touch(exist_ok=True)  # Create if it doesn't exist

        self.states_to_ignore_list = ["MESSAGE", "FINISH_OPERATION"]

        # Set default state
        self._record_state_change(self.nextion_state.state)

        self._state_recorded = False
        self.__exit = True

        logger.info("StateChangeRecorder initialized...")

    def _record_state_change(self, new_state: NextionPagingState):
        """
            Writes the current state to the state file
        @return:
        """
        state_dict = dict()

        state = str(new_state).split(' ')
        state = state[len(state)-1]
        # logger.info(f"State to record: {state}")

        if state in self.states_to_ignore_list:
            logger.info(f"Ignored state change {state}")
        else:
            state_dict["STATE"] = state
            with open(self._state_file, "w") as f:
                json.dump(state_dict, f)
            logger.info(f"State [{state}] was stored to {self._state_file}")

        self._state_recorded = False

    def exit(self):
        self.__exit = True

    def run(self):
        self.__exit = False

        while not self.__exit:

            with self.nextion_interface.lock:

                if self.nextion_interface.state_change and not self._state_recorded:
                    self.nextion_interface.clear_state_change()
                    self._state_recorded = True
                    from_state = self.nextion_state.p_state
                    to_state = self.nextion_state.state
                    # logger.info(f"State changed from [{from_state}] to [{to_state}]")
                    self._record_state_change(new_state=to_state)

            sleep(0.009)  # Should not be divisible by 100ms


def get_device_state():
    # Make sure the PAGING_MASTER_TMP_PATH exists
    if os.path.exists(PAGING_MASTER_TMP_PATH):
        device_state_file = os.path.join(PAGING_MASTER_TMP_PATH, STATE_FILE)
        try:
            fact_info_json = read_json(device_state_file)
            device_state = fact_info_json["STATE"]
            return device_state
        except Exception as e:
            logger.error("Exception found: {}".format(e))
    return ""
