from nextion.nextion_enum import NextionPagingState, KeyID, KeyStatus
from nextion.utils import BitManip
from utils import get_logger, get_uci

logger = get_logger(name="StateVariables",
                    log_file="/var/log/state_variables_log")


class NextionStateVariable:
    def __init__(
        self,
        current_state: NextionPagingState = NextionPagingState.IDLE,
        previous_state: NextionPagingState = NextionPagingState.IDLE,
    ):
        self.__current_state = current_state
        self.__previous_state = NextionPagingState(previous_state)
        self.__state_changed = True

    def __repr__(self):
        return f"Nextion [p_state -> state] : {self.p_state.name} -> {self.state.name}"

    @property
    def state_changed(self):
        return self.__state_changed

    def state_synced(self):
        self.__state_changed = False

    @property
    def state(self):
        return self.__current_state

    @state.setter
    def state(self, next_state: NextionPagingState):
        self.__state_changed = True
        self.__previous_state = NextionPagingState(self.__current_state)
        self.__current_state = next_state

    @property
    def p_state(self):
        return self.__previous_state


class ButtonState:
    """This class controls the state of buttons on nextion panel."""

    def __init__(self):
        self._number_pages = int(get_uci("icpaging-master", "main", "panel_pages"))
        self._relay_button_id = (self._number_pages * 48) + 64

        # LED variables
        self.__leds1 = 0
        self.__leds2 = 0
        self.__leds3 = 0
        self.__leds4 = 0
        # LED flags
        # These flags are reset everytime respective led is accessed.
        # They are set when an LED is changed.
        self.__led1_changed = True
        self.__led2_changed = True
        self.__led3_changed = True
        self.__led4_changed = True

        if self._number_pages >= 1:
            self.__leds5 = 0
            self.__leds6 = 0
            self.__leds7 = 0

            self.__led5_changed = True
            self.__led6_changed = True
            self.__led7_changed = True

        self.reset()
        # Only add buttons that have been used
        self._buttons_status_dict = {}  # KeyID : Status

    def reset(self):
        self.__leds1 = 0
        self.__leds2 = 0
        self.__leds3 = 0
        self.__leds4 = 0

        self.__led1_changed = True
        self.__led2_changed = True
        self.__led3_changed = True
        self.__led4_changed = True

        if self._number_pages >= 1:
            self.__leds5 = 0
            self.__leds6 = 0
            self.__leds7 = 0

            self.__led5_changed = True
            self.__led6_changed = True
            self.__led7_changed = True

    def reset_station_selection(self):
        """This command resets all the stations."""
        for key in KeyID:
            if "STATION" in key.name:
                # self.reset_button(key.value)  # Used to be this
                self.reset_button(key)
        logger.info("Stations reset")

    def reset_group_selection(self):
        for key in KeyID:
            if "GROUP" in key.name:
                # self.reset_button(key.value)  # Used to be this
                self.reset_button(key)
        logger.info("Groups reset")

    def set_button(self, button_id):
        if -1 < button_id < 16:
            self.__led1_changed = True
            self.__leds1 = BitManip.set_bit(self.__leds1, button_id)
        elif 15 < button_id < 32:
            self.__led2_changed = True
            self.__leds2 = BitManip.set_bit(self.__leds2, button_id - 16)
        elif 31 < button_id < 48:
            self.__led3_changed = True
            self.__leds3 = BitManip.set_bit(self.__leds3, button_id - 32)
        elif 47 < button_id < 64:
            self.__led4_changed = True
            self.__leds4 = BitManip.set_bit(self.__leds4, button_id - 48)
        elif 63 < button_id < 80 and (button_id != self._relay_button_id):
            self.__led5_changed = True
            self.__leds5 = BitManip.set_bit(self.__leds5, button_id - 64)
        elif 79 < button_id < 96:
            self.__led6_changed = True
            self.__leds6 = BitManip.set_bit(self.__leds6, button_id - 80)
        elif 95 < button_id < 112:
            self.__led7_changed = True
            self.__leds7 = BitManip.set_bit(self.__leds7, button_id - 96)

        elif button_id == self._relay_button_id:
            # relay button
            pass
        else:
            logger.warning("Failed SET")
            raise Exception(f"Button id exceeded the limit {button_id}")

        self.set_button_status(button_id, KeyStatus.PRESSED)  # This will be only reached if it was a valid button ID
        logger.info("Done")

    def flip_button(self, button_id):
        if -1 < button_id < 16:
            self.__led1_changed = True
            self.__leds1 = BitManip.flip_bit(self.__leds1, button_id)
        elif 15 < button_id < 32:
            self.__led2_changed = True
            self.__leds2 = BitManip.flip_bit(self.__leds2, button_id - 16)
        elif 31 < button_id < 48:
            self.__led3_changed = True
            self.__leds3 = BitManip.flip_bit(self.__leds3, button_id - 32)
        elif 47 < button_id < 64:
            self.__led4_changed = True
            self.__leds4 = BitManip.flip_bit(self.__leds4, button_id - 48)
        elif 63 < button_id < 80 and (button_id != self._relay_button_id):
            self.__led5_changed = True
            self.__leds5 = BitManip.flip_bit(self.__leds5, button_id - 64)
        elif 79 < button_id < 96:
            self.__led6_changed = True
            self.__leds6 = BitManip.flip_bit(self.__leds6, button_id - 80)
        elif 95 < button_id < 112:
            self.__led7_changed = True
            self.__leds7 = BitManip.flip_bit(self.__leds7, button_id - 96)
        elif button_id == self._relay_button_id:
            # relay button
            pass
        else:
            logger.warning("Failed FLIP")
            raise Exception(f"Button id exceeded the limit {button_id}")

        # This will be only reached if it was a valid button ID
        current_status = self.get_button_status(button_id)
        # logger.info("Current status: %s", current_status)
        if current_status == KeyStatus.RELEASED:
            new_status = KeyStatus.PRESSED
        else:
            new_status = KeyStatus.RELEASED
        self.set_button_status(button_id, new_status)
        logger.info("Done")

    def reset_button(self, button_id):
        if -1 < button_id < 16:
            self.__led1_changed = True
            self.__leds1 = BitManip.reset_bit(self.__leds1, button_id)
        elif 15 < button_id < 32:
            self.__led2_changed = True
            self.__leds2 = BitManip.reset_bit(self.__leds2, button_id - 16)
        elif 31 < button_id < 48:
            self.__led3_changed = True
            self.__leds3 = BitManip.reset_bit(self.__leds3, button_id - 32)
        elif 47 < button_id < 64:
            self.__led4_changed = True
            self.__leds4 = BitManip.reset_bit(self.__leds4, button_id - 48)
        elif 63 < button_id < 80 and (button_id != self._relay_button_id):
            self.__led5_changed = True
            self.__leds5 = BitManip.reset_bit(self.__leds5, button_id - 64)
        elif 79 < button_id < 96:
            self.__led6_changed = True
            self.__leds6 = BitManip.reset_bit(self.__leds6, button_id - 80)
        elif 95 < button_id < 112:
            self.__led7_changed = True
            self.__leds7 = BitManip.reset_bit(self.__leds7, button_id - 96)

        elif button_id == self._relay_button_id:
            # relay button
            pass
        else:
            logger.warning("Failed RESET")
            raise Exception(f"Button id exceeded the limit {button_id}")

        self.set_button_status(button_id, KeyStatus.RELEASED)  # This will be only reached if it was a valid button ID

    def get_state(self, button_id):
        if -1 < button_id < 16:
            return BitManip.get_bit(self.__leds1, button_id)
        elif 15 < button_id < 32:
            return BitManip.get_bit(self.__leds2, button_id - 16)
        elif 31 < button_id < 48:
            return BitManip.get_bit(self.__leds3, button_id - 32)
        elif 47 < button_id < 64:
            return BitManip.get_bit(self.__leds4, button_id - 48)
        elif 63 < button_id < 80 and (button_id != self._relay_button_id):
            return BitManip.get_bit(self.__leds5, button_id - 64)
        elif 79 < button_id < 96:
            return BitManip.get_bit(self.__leds6, button_id - 80)
        elif 95 < button_id < 112:
            return BitManip.get_bit(self.__leds7, button_id - 96)

        elif button_id == self._relay_button_id:
            # relay button
            return 0
        else:
            logger.warning("Failed GET")
            raise Exception(f"Button id exceeded the limit {button_id}")

    def remove_released_buttons(self):
        """
            Remove the buttons which have their status as RELEASED.
            Call this function only in IDLE state. Currently, not used.
        @return:
        """

        logger.info("Dict before removing RELEASED buttons: %s", self.get_buttons_status_dict)
        for key in self._buttons_status_dict:
            value = self._buttons_status_dict[key]
            logger.info("Current dict Key: %s", key)
            logger.info("Current dict Value: %s", value)

            if value == KeyStatus.RELEASED:
                logger.warning("Removing Key: %s", key)
                self._buttons_status_dict.pop(key)

        logger.info("Dict after removing RELEASED buttons: %s", self.get_buttons_status_dict)

    def set_button_status(self, key_id: KeyID, status: KeyStatus = KeyStatus.PRESSED):
        # logger.info("Key ID to check: %s", key_id)

        if key_id in KeyID:
            # logger.info("ID found in KeyID")
            self._buttons_status_dict[key_id.name] = status
            # logger.info("Current dictionary: %s", self.get_buttons_status_dict)
        else:
            raise Exception("Invalid KeyID")

    def get_button_status(self, key_id: KeyID):

        if key_id.name in self._buttons_status_dict.keys():
            return self._buttons_status_dict[key_id.name]
        else:
            # logger.warning(f"Key {key_id} status not found. Using KeyID.RELEASED as default")
            return KeyStatus.RELEASED

    @property
    def get_buttons_status_dict(self):
        return self._buttons_status_dict

    @property
    def led1_changed(self):
        return self.__led1_changed

    @property
    def led2_changed(self):
        return self.__led2_changed

    @property
    def led3_changed(self):
        return self.__led3_changed

    @property
    def led4_changed(self):
        return self.__led4_changed

    @property
    def led5_changed(self):
        return self.__led5_changed

    @property
    def led6_changed(self):
        return self.__led6_changed

    @property
    def led7_changed(self):
        return self.__led7_changed

    @property
    def leds1(self) -> bytes:
        self.__led1_changed = False
        return f"leds1={self.__leds1}".encode()

    @property
    def leds2(self) -> bytes:
        self.__led2_changed = False
        return f"leds2={self.__leds2}".encode()

    @property
    def leds3(self) -> bytes:
        self.__led3_changed = False
        return f"leds3={self.__leds3}".encode()

    @property
    def leds4(self) -> bytes:
        self.__led4_changed = False
        return f"leds4={self.__leds4}".encode()

    @property
    def leds5(self) -> bytes:
        self.__led5_changed = False
        return f"leds5={self.__leds5}".encode()

    @property
    def leds6(self) -> bytes:
        self.__led6_changed = False
        return f"leds6={self.__leds6}".encode()

    @property
    def leds7(self) -> bytes:
        self.__led7_changed = False
        return f"leds7={self.__leds7}".encode()
