import argparse
import serial
from os import path
import logging
import logging.config
import time



class NextionFWUpdater:
    def __init__(
        self, file_path: str, baud_rate: int, serial_port: str, timeout: int = 0.1
    ):
        logging.config.fileConfig("logging.conf")
        self.logger = logging.getLogger("NextionFWUpdate")
        self.__file_path = file_path
        self.__baud_rate: int = baud_rate
        self.__serial_port: int = serial_port
        self.__timeout: int = timeout
        self.__file_size = path.getsize(file_path)
        self.__packet_size = 4096
        self.__baud_com = 9200

    def firmware_file(self):
        """Generator for the Nextion TFT file. Reads 4096 bytes from the memory."""
        with open(self.__file_path, "rb") as f:
            for i in range(0, self.__file_size, self.__packet_size):
                data = f.read(self.__packet_size)
                yield data

    def initialize_connection(self, ser):
        self.logger.info("Initializing FW update connection to the Nextion Panel.")
        ser.write(b"\xff\xff\xff")
        ser.write(b"connect")
        ser.write(b"\xff\xff\xff")
        time.sleep(0.5)
        ser.write(f"whmi-wri {self.__file_size},{self.__baud_rate},res0".encode())
        ser.write(b"\xff\xff\xff")
        time.sleep(0.5)
        self.wait_until_valid_panel_response(ser, b"comok")

    def wait_until_valid_panel_response(self, ser, response):
        nextion_response = ser.read(128)
        while True:
            if response in nextion_response:
                return
            else:
                nextion_response = ser.read(128)
                time.sleep(0.1)

    def upload_file(self):
        with serial.Serial(
            self.__serial_port, self.__baud_rate, timeout=self.__timeout
        ) as ser:
            current_fw_version = self.get_current_version(ser)
            new_fw_version = self.get_new_version(self.__file_path)
            self.logger.info(f"The current fw version is: {current_fw_version}")
            self.logger.info(f"The new fw version is: {new_fw_version}")

            if current_fw_version != new_fw_version:
                self.initialize_connection(ser)
                self.logger.info("Uploading TFT file.")
                for data in self.firmware_file():
                    ser.write(data)
                    self.wait_until_valid_panel_response(ser, b"\x05")
                # time.sleep(5) # it takes 3-4 seconds to restart a device after update.
            else:
                message = "FW versions are the same. No need to update the panel."
                print(message)
                self.logger.info(message)

    @staticmethod
    def get_current_version(ser) -> str:
        terminating_characters = b"\xff\xff\xff"
        ser.write(b"get version.txt" + terminating_characters)
        time.sleep(0.5)
        data = ser.read_until(terminator=terminating_characters)
        print(data)
        current_fw = 'unknown'
        if data[0] == ord("\x70"):
            current_fw = data[1:-len(terminating_characters)]
        print(current_fw)
        # self.logger.info(f"The version is {data}")
        return current_fw.decode()

    # TODO: Extract the new FW version from the filename
    @staticmethod
    def get_new_version(file_path):
        paths, filename = path.split(file_path)
        no_ext_file = filename.rsplit(".", 1)[0]
        new_version = no_ext_file.split('-')[-1:]
        return new_version[0]


def parse_args():
    parser = argparse.ArgumentParser(
        description="Script to upload a firmware to Nextion panel"
    )
    parser.add_argument("--baud_rate", type=int, default=115200)
    parser.add_argument("--port", type=str, default="/dev/ttyUSB0")
    parser.add_argument(
        "--file", type=str, required=True, help="path of the new nextion firmware"
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()
    nextion_updater = NextionFWUpdater(args.file, args.baud_rate, args.port)
    nextion_updater.upload_file()
