import time
from datetime import datetime
from threading import Thread

from master.paging_master import PagingMaster
from nextion.nextion_enum import NextionPagingState, KeyID, KeyStatus
from nextion.nextion_interface import NextionInterface
from nextion.nextion_panel import NextionPanel
from nextion.services.state_change_recorder import StateChangeRecorder, get_device_state

from server.services.ring_service import RingService
from server.services.record_service import MessageRecordService
from server.services.recorded_message_service import RecordedMessagesService
from server.services.sound_level_updater import SoundLevelUpdater, get_sound_level
from server.services.sync_files_wd import SyncFilesWD

from server.utils import audio_uci_to_config_dict

from utils import get_logger, mcu_gpio_control, NextionPanelReset

logger = get_logger("PagingServer")


class PagingServer:
    # Added "nextion_panel: NextionPanel" to allow for manual buttons state changes
    def __init__(self, nextion_panel: NextionPanel, nextion_interface: NextionInterface):
        try:
            config = audio_uci_to_config_dict()
        except Exception as e:
            logger.warning(f"Could not load settings: {e}")
        self._paging_priority = config["paging_priority"]
        self.__pmaster = PagingMaster()
        self.__sound_level_updater = SoundLevelUpdater()  # Used when recording or playing recorded files
        self._nextion_panel = nextion_panel
        self.nextion_interface = nextion_interface
        self.nextion_state = nextion_interface.nextion_state
        try:
            self.button_state = nextion_interface.buttons_state
        except Exception as e:
            logger.warning(f"Error: {e}")

        # Service accessors
        self.__record_service = None
        self.__ring_service = None
        self.__recorded_message_service = None
        self.__sync_files_wd = None
        self.__state_change_recorder = None
        self.__initialise_services(nextion_panel, nextion_interface, self.__sound_level_updater)

        # Thread handles
        self.__status_listen_service_thread = None
        self.__status_send_thread = None
        self.__monitor_calling_client_thread = None
        self.__paging_thread = None
        self.__call_thread = None
        self.__state_change_recorder_thread = None
        self.__sync_files_wd_thread = None
        self.__send_message_with_timeout_thread = None

        # Flags
        self.__worker_threads_list = []  # If there are multiple worker threads to be closed, add them to the list
        self.__kill_thread_flag = False  # Only set True from __close_worker_threads function
        self.__start_away_answer_flag = False  # Set True from check_clients when an incoming call is received in UNATTENDED
        self.away_press_start_timer = None  # Use to mark the start of an Away button press.
        self._change_nextion_message = True  # Use to indicate when the message needs to be changed

    def __initialise_services(self, nextion_panel: NextionPanel, nextion_interface: NextionInterface,
                              sound_level_updater: SoundLevelUpdater):
        """
            Initialise all the different services, necessary for the app operation
        @param nextion_panel:
        @param nextion_interface:
        @return: None
        """
        self.__record_service = MessageRecordService(nextion_panel, nextion_interface, sound_level_updater)
        self.__ring_service = RingService()
        self.__recorded_message_service = RecordedMessagesService(nextion_panel, nextion_interface, sound_level_updater)
        # self.__sync_files_wd = SyncFilesWD(["*.wav"])
        self.__state_change_recorder = StateChangeRecorder(nextion_interface)
        logger.info("All services initialised.")

    def __run_services(self):
        """
            Runs all the different services. Only call at the start from self.run()
        @return: None
        """
        self._monitor_calling_client(start=True)
        self._listen_status_message_service(start=True)
        # self._sync_files_service(start=True)
        self._state_change_recorder_service(start=True)
        logger.info("All services started.")

    def __stop_services(self):
        """
            Properly stop each individual service
        @return:
        """
        self._status_send_service(start=False)
        self._listen_status_message_service(start=False)
        self._monitor_calling_client(start=False)
        self._incoming_call(start=False)
        self._sync_files_service(start=False)
        self._state_change_recorder_service(start=False)
        self._nextion_panel.stop()
        logger.info("All services stopped.")

    # TODO: Not fully implemented and tested yet
    def _state_change_recorder_service(self, start=True):
        if start:
            if self.__state_change_recorder_thread is None:
                try:
                    logger.info("Starting StateChangeRecorder service")
                    self.__state_change_recorder_thread = Thread(
                        target=self.__state_change_recorder.run
                    )
                    self.__state_change_recorder_thread.daemon = True
                    self.__state_change_recorder_thread.setName("Thread-StateChangeRecorder")
                    self.__state_change_recorder_thread.start()
                    logger.info("Done")
                except Exception as e:
                    logger.error(f"Couldn't start the StateChangeRecorder thread: {e}")
            else:
                logger.warning(
                    "Attempt to start the StateChangeRecorder service. Service is already running!"
                )
        else:
            if self.__state_change_recorder_thread is not None:
                self.__state_change_recorder.exit()
                self.__state_change_recorder_thread.join()
                self.__state_change_recorder_thread = None
                logger.info("Stopped _state_change_recorder_service")

    # TODO: Not fully implemented and tested yet
    def _sync_files_service(self, start=True):
        if start:
            if self.__sync_files_wd_thread is None:
                try:
                    self.__sync_files_wd_thread = Thread(
                        target=self.__sync_files_wd.run
                    )
                    self.__sync_files_wd_thread.daemon = True
                    self.__sync_files_wd_thread.setName("Thread-SyncFilesWD")
                    self.__sync_files_wd_thread.start()
                except Exception as e:
                    logger.error(f"Couldn't start the SyncFilesWD thread: {e}")
            else:
                logger.warning(
                    "Attempt to start the SyncFilesWD. Service is already running!"
                )
        else:
            if self.__sync_files_wd_thread is not None:
                self.__sync_files_wd.exit()
                self.__sync_files_wd_thread.join()
                self.__sync_files_wd_thread = None
                logger.info("Stopped _sync_files_service")

    def _status_send_service(self, start=True):
        if start:
            if self.__status_send_thread is None:
                logger.info("Send status service started.")
                self.__status_send_thread = Thread(
                    target=self.__pmaster.send_status_message
                )
                self.__status_send_thread.daemon = True
                self.__status_send_thread.setName("Thread-StatusSend")
                self.__status_send_thread.start()
            else:
                logger.warning(
                    "Attempt to start the send status service. Service is already running!"
                )
        else:
            if self.__status_send_thread is not None:
                logger.info("Stopped _status_send_service")
                self.__status_send_thread.join()

    def _listen_status_message_service(self, start=True):
        if start:
            if self.__status_listen_service_thread is None:
                logger.info("Status listening service started")
                self.__status_listen_service_thread = Thread(
                    target=self.__pmaster.listen_client_status
                )
                self.__status_listen_service_thread.daemon = True
                self.__status_listen_service_thread.setName("Thread-StatusListen")
                self.__status_listen_service_thread.start()
            else:
                logger.warning(
                    "Attempt to start the LISTEN status service. Service is already running!"
                )
        else:
            if self.__status_listen_service_thread is not None:
                self.__status_listen_service_thread.join()
                logger.info("Stopped _listen_status_message_service")

    def check_clients(self):
        while not self.__pmaster.exit_flag:
            with self.nextion_interface.lock:
                self.__pmaster.purge_stale_calling_client()
                if len(self.__pmaster.calling_client) != 0:

                    if self.nextion_state.state == NextionPagingState.IDLE:
                        self.nextion_state.state = NextionPagingState.INCOMING_CALL_REQUEST
                        self.nextion_interface.message = f"Station {self.__pmaster.calling_client['id']} calling"
                        # TODO: handle the calling station here.
                        self.__ring_service.active_station = self.__pmaster.calling_client["id"]

                    # TODO: Handle away and auto rejects here.
                    elif (
                            self.__pmaster.in_call and
                            self.nextion_state.state == NextionPagingState.IN_ACTIVE_CALL and
                            self.__pmaster.calling_client['id'] != self.__pmaster.client_number_in_call
                    ):
                        self.__pmaster.reject_call(reason=0)
                        message = f"Busy in another call.\r\n" \
                            f"Call from [{self.__pmaster.calling_client['id']}] rejected"
                        self.nextion_interface.message = message
                        logger.warning(message)
                        time.sleep(1)  # TODO: Find a better way to change back the message without sleep
                        message = f"In Active call"
                        self.nextion_interface.message = message

                    elif self.nextion_state.state == NextionPagingState.UNATTENDED_MODE:
                        message = f"Station {self.__pmaster.calling_client['id']} calling"
                        self.nextion_interface.message = message
                        if not self.__start_away_answer_flag:
                            self.__start_away_answer_flag = True

                    elif self.nextion_state.state == NextionPagingState.CALL_FORWARDING_MODE:
                        self.__pmaster.reject_call(reason=2)
                        self.nextion_interface.message = "Call Rejected"
                        self._change_nextion_message = True

                else:
                    if self.nextion_state.state == NextionPagingState.INCOMING_CALL_REQUEST:
                        self.nextion_state.state = NextionPagingState.IDLE
            time.sleep(0.5)

    def _monitor_calling_client(self, start: bool):
        if start:
            if self.__monitor_calling_client_thread is None:
                self.__monitor_calling_client_thread = Thread(target=self.check_clients)
                self.__monitor_calling_client_thread.daemon = True
                self.__monitor_calling_client_thread.setName("Thread-MonitorCallingClient")
                self.__monitor_calling_client_thread.start()
                logger.info("Monitoring the clients for incoming call")
            else:
                logger.warning(
                    "Attempt to start the LISTEN status service. Service is already running!"
                )
        else:
            if self.__monitor_calling_client_thread is not None:
                self.__monitor_calling_client_thread.join()

    def _page(self, start: bool):
        stop = not start
        if start:
            if self.__paging_thread is None:
                clients = self.nextion_interface.selected_clients
                if len(clients):
                    logger.debug(f"Paging to {clients}")
                    self.__paging_thread = Thread(
                        target=self.__pmaster.send_page_group_request,
                        kwargs={"receiver_id": clients, "priority": int(self._paging_priority)},
                    )
                    self.__paging_thread.daemon = True
                    self.__paging_thread.setName("Thread-Paging")
                    self.__paging_thread.start()
                    self.nextion_interface.message = f"Paging to Station {clients}"
                else:
                    self.nextion_interface.message = f"Paging: No Station selected!"
                    logger.debug("No clients selected")
            else:
                if self.__pmaster.capture_disabled:
                    self.nextion_interface.message = "Paging, microphone disabled!"
        elif stop and self.__paging_thread is not None:
            clients = self.nextion_interface.selected_clients
            logger.info(f"Paging finished for clients {clients}")
            self.__pmaster.end_paging(clients)
            if self.__paging_thread is not None:
                self.__paging_thread.join()
                self.__paging_thread = None
        else:
            logger.error("Unknown paging state!")

    # TODO: Not fully tested
    def _page_custom_message(self, start: bool):
        stop = not start
        if start:
            # Use the same thread, as paging and sending message at the same time is not a valid use case
            if self.__paging_thread is None:
                clients = self.nextion_interface.selected_clients
                if len(clients):
                    file_path = self.__recorded_message_service.get_message("custom_message")
                    if file_path != "No file found":
                        logger.debug(f"Paging custom message to {clients}")
                        self.__paging_thread = Thread(
                            target=self.__pmaster.send_page_group_request,
                            kwargs={"receiver_id": clients, "priority": int(self._paging_priority),
                                    "audio_file_path": file_path},
                        )
                        self.__paging_thread.daemon = True
                        self.__paging_thread.setName("Thread-PageCustomMsg")
                        self.__paging_thread.start()
                        message = f"Paging custom message to Station {clients}"
                        self.nextion_interface.message = message
                    else:
                        message = "No custom message file found"
                        logger.warning(message)
                        self.nextion_interface.message = message

                else:
                    self.nextion_interface.message = f"Paging: No Station selected!"
                    logger.warning("No clients selected")
            else:
                if self.__pmaster.capture_disabled:
                    self.nextion_interface.message = "Paging, microphone disabled!"
                # TODO
                elif self.__pmaster.end_message_transmission():
                    message = f"Changing state to FINISH_OPERATION"
                    self.nextion_interface.message = message
                    self.nextion_state.state = NextionPagingState.FINISH_OPERATION

        elif stop:
            if self.__paging_thread is not None:
                clients = self.nextion_interface.selected_clients
                logger.info(f"Paging finished for clients {clients}")
                self.__pmaster.end_paging(clients)
                if self.__paging_thread is not None:
                    self.__paging_thread.join()
                    self.__paging_thread = None
            else:
                logger.warning("Paging thread already closed")
        else:
            logger.error("Unknown paging state!")

    def _accept_call(self):
        if len(self.__pmaster.calling_client) != 0:
            self.__pmaster.send_call_accept_message()
            if self.nextion_interface.is_half_duplex:  # HALF DUPLEX
                if self.__start_away_answer_flag:
                    # TODO
                    logger.warning("THIS HAS NOT BEEN YET IMPLEMENTED")
                    self.__call_thread = Thread(target=self.__pmaster.half_duplex_call,
                                                kwargs={"mode": "UNATTENDED",
                                                        "filepath": self.__recorded_message_service.get_message(
                                                            "away_message")}
                                                )
                    self.nextion_interface.message = "Sending away message HALF duplex"
                else:
                    self.__call_thread = Thread(target=self.__pmaster.half_duplex_call)
                    message = "In Active call (Listen)"
                    logger.info(message)
                    self.nextion_interface.message = message
            else:  # FULL DUPLEX
                if self.__start_away_answer_flag:
                    self.__call_thread = Thread(
                        target=self.__pmaster.full_duplex_call,
                        kwargs={"mode": "UNATTENDED",
                                "filepath": self.__recorded_message_service.get_message("away_message")}
                    )
                    message = "Sending away message FULL duplex to" \
                              "\r\nClient {}".format(self.__pmaster.active_call_client["id"])
                    # logger.info(message)
                    self.nextion_interface.message = message
                else:
                    self.__call_thread = Thread(target=self.__pmaster.full_duplex_call)
                    self.nextion_interface.message = "In Active call"
            self.__call_thread.daemon = True
            self.__call_thread.setName("Thread-Call")
            self.__call_thread.start()
            self.nextion_interface.update_call_info(
                self.__pmaster.active_call_client["id"],
                self.__pmaster.active_call_client["ip"],
            )
        else:
            self.nextion_interface.message = "No incoming call!"
            logger.warning("No active client present")

    def _toggle_half_duplex_mode(self):
        if (
            self.nextion_interface.is_half_duplex_listen_mode
            and not self.__pmaster.is_listen_mode
        ):
            self.nextion_interface.message = "In Active call (Listen)"
            self.__pmaster.half_duplex_listen_mode()
        elif (
            not self.nextion_interface.is_half_duplex_listen_mode
            and self.__pmaster.is_listen_mode
        ):
            self.nextion_interface.message = "Active call (Talk)"
            self.__pmaster.half_duplex_talk_mode()

    def _incoming_call(self, start: bool):
        stop = not start
        # logger.warning(f"_incoming_call, start: {start}")
        if start:
            if self.__call_thread is None:
                self._accept_call()
            elif self.nextion_interface.is_half_duplex:
                self._toggle_half_duplex_mode()
            if self.nextion_interface.relay:
                self.nextion_interface.message += ": Door open!"
                self.__pmaster.send_relay_message(
                    receiver_id=self.__pmaster.active_call_client["id"],
                    relay_number=5
                )
        elif stop and self.__call_thread is not None:
            # End the call gracefully.
            self.nextion_interface.message = "Hanging up"
            self.__pmaster.close_call()
            if self.__call_thread is not None:
                self.__call_thread.join()
                self.__call_thread = None
            self.nextion_interface.message = "Call closed"

    def __close_worker_threads(self):
        if self.__send_message_with_timeout_thread is not None:
            self.__kill_thread_flag = True
            # logger.info("Closing %s...", self.__send_message_with_timeout_thread.name)
            self.__send_message_with_timeout_thread.join()
            self.__send_message_with_timeout_thread = None
            # logger.info("Done!")

            self.__kill_thread_flag = False  # Reset back to default
            del self.__worker_threads_list[:]  # reset the list
            logger.info("Worker threads closed")
        else:
            # logger.info("No worker threads")
            pass

    def __start_worker_threads(self, message, timeout):
        if self.__send_message_with_timeout_thread is None:
            self.__send_message_with_timeout_thread = Thread(
                target=self._send_message_with_timeout,
                kwargs={"message": message, "timeout": timeout}
            )
            self.__send_message_with_timeout_thread.daemon = True
            self.__send_message_with_timeout_thread.setName("Thread-SendTimeoutMsg")
            self.__send_message_with_timeout_thread.start()
            self.__worker_threads_list.append(self.__send_message_with_timeout_thread)

    # Opened in a separate thread to send a message after a timeout has expired
    def _send_message_with_timeout(self, message, timeout):
        count = 0
        delay = 0.1  # Check if the thread should be killed earlier
        while self.nextion_state.state != NextionPagingState.FINISH_OPERATION:
            # Break when needed
            if self.__kill_thread_flag:
                break

            if count % (timeout // delay) == 0:
                self.nextion_interface.message = message

            count += 1
            time.sleep(delay)

    def run(self):

        self.__run_services()
        ready_msg = True
        time.sleep(0.5)  # Give some time to the screen to get responsive
        while True:
            with self.nextion_interface.lock:
                if self._nextion_panel.restart_main_app_flag:
                    self.nextion_interface.message = "Please wait while the Nextion panel\r\nis properly reset..."
                    raise NextionPanelReset
                if ready_msg:
                    ready_msg = False
                    msg = "Ready"
                    self.nextion_interface.message = msg
                # logger.info("Nextion panel state: %s", self.nextion_state.state)
                if self.nextion_state.state == NextionPagingState.PAGING_ACTIVE:
                    self._page(start=True,)
                elif self.nextion_state.state == NextionPagingState.IN_ACTIVE_CALL:
                    if self.__ring_service.is_running:
                        self.__ring_service.exit()

                    # Needs to be here, so half duplex can work properly and relay messages can be detected
                    # TODO: Find a more efficient way of doing that
                    self._incoming_call(start=True)

                    if not self.__pmaster.in_call:
                        # close call when mic is disabled or connection from client is lost.
                        self._incoming_call(start=False)

                        if (
                            self.__pmaster.capture_disabled
                            and self.__pmaster.client_lost
                        ):
                            self.nextion_interface.message = (
                                "Hanging up, Microphone disabled and connection lost!"
                            )
                        elif self.__pmaster.client_lost:
                            self.nextion_interface.message = (
                                "Hanging up, connection lost!"
                            )
                        elif self.__pmaster.capture_disabled:
                            self.nextion_interface.message = (
                                "Hanging up, Microphone disabled!"
                            )
                        else:
                            self.nextion_interface.message = "Hanging up"
                        self.nextion_state.state = NextionPagingState.IDLE
                elif self.nextion_state.state == NextionPagingState.IDLE:

                    if self.__send_message_with_timeout_thread is not None:
                        self.__close_worker_threads()  # Make sure the messages were stopped

                    if self.__paging_thread is not None:
                        self._page(start=False)
                    elif self.__call_thread is not None:
                        self._incoming_call(start=False)
                    elif self.nextion_state.p_state == NextionPagingState.INCOMING_CALL_REQUEST:
                        # This case is possible when the call was rejected by the panel.
                        self.__ring_service.exit()
                        self.__pmaster.reject_call(reason=1)
                        self.nextion_interface.message = "Call Rejected"
                        self.nextion_state.state = NextionPagingState.IDLE

                    # elif self.nextion_state.p_state == NextionPagingState.IN_ACTIVE_CALL and \
                    #         self.__call_thread is not None:
                    #     self._incoming_call(start=False)

                    elif self.nextion_state.p_state == NextionPagingState.MESSAGE:
                        self.__record_service.exit()

                    elif self.nextion_state.p_state == NextionPagingState.UNATTENDED_MODE or \
                            self.nextion_state.p_state == NextionPagingState.CALL_FORWARDING_MODE:
                        self.__close_worker_threads()  # Taken care at entering IDLE state
                        if self.away_press_start_timer is not None:
                            self.away_press_start_timer = None
                        if not self._change_nextion_message:
                            self._change_nextion_message = True

                elif self.nextion_state.state == NextionPagingState.INCOMING_CALL_REQUEST:
                    self.__ring_service.run()
                    logger.debug("Incoming call!")

                # TODO: This state is set when holding the Away button for more than 4 seconds.
                #    This happens inside UNATTENDED mode state.
                elif self.nextion_state.state == NextionPagingState.CALL_FORWARDING_MODE:
                    if self.away_press_start_timer is not None:
                        self.away_press_start_timer = None

                    if self._change_nextion_message:
                        self._change_nextion_message = False
                        self.nextion_interface.message = "All incoming calls will be rejected!!!"
                        # self.__start_worker_threads(message="All incoming calls will be rejected!!!", timeout=2)

                elif self.nextion_state.state == NextionPagingState.UNATTENDED_MODE:

                    if self.away_press_start_timer is None:
                        self.away_press_start_timer = datetime.now()

                    # Only start it if a call is not ongoing
                    if not self.__pmaster.in_call and self._change_nextion_message:
                        self._change_nextion_message = False
                        self.nextion_interface.message = "Automatic answer to calls..."

                    # Detect a long button press
                    if self.button_state.get_button_status(KeyID.AWAY_BUTTON) == KeyStatus.PRESSED:
                        if (datetime.now() - self.away_press_start_timer).seconds >= 4:
                            self._change_nextion_message = True
                            self.__close_worker_threads()
                            logger.info("This was a long press. Changing to CALL_FORWARDING_MODE")
                            self.nextion_state.state = NextionPagingState.CALL_FORWARDING_MODE

                    # Set inside check_clients()
                    if self.__start_away_answer_flag:
                        self.__close_worker_threads()
                        if not self.__pmaster.in_call:
                            logger.info("Accepting the call")
                            self._incoming_call(start=True)

                        # # Play locally
                        # if not self.__recorded_message_service.is_running:
                        #     # self.__close_worker_threads()
                        #     self.__start_worker_threads(message="Answer with away message...", timeout=1)
                        #     self.__recorded_message_service.play_locally(audio_file_type="away_message",
                        #                                                  audio_file_id="")
                        # # Only returns True after the local playback thread has finished
                        # if self.__recorded_message_service.exit():
                        #     logger.warning("Local playback has finished for UNATTENDED MODE.")
                        #     # self.nextion_state.state = NextionPagingState.FINISH_OPERATION

                        if self.__pmaster.end_message_transmission or not self.__pmaster.in_call:
                            self.__start_away_answer_flag = False
                            message = f"Changing state to FINISH_OPERATION"
                            logger.info(message)
                            # close call when mic is disabled or connection from client is lost.
                            self._incoming_call(start=False)

                            if self.__pmaster.capture_disabled and self.__pmaster.client_lost:
                                message = "Hanging up, Microphone disabled and connection lost!"
                            elif self.__pmaster.client_lost:
                                message = "Hanging up, connection lost!"
                            elif self.__pmaster.capture_disabled:
                                message = "Hanging up, Microphone disabled!"
                            else:
                                message = "Hanging up"
                            self.nextion_interface.message = message
                            self.nextion_state.state = NextionPagingState.FINISH_OPERATION

                elif self.nextion_state.state == NextionPagingState.FINISH_OPERATION:
                    if (
                        self.nextion_state.p_state == NextionPagingState.RECORDING_AWAY_MESSAGE or
                        self.nextion_state.p_state == NextionPagingState.RECORDING_CUSTOM_MESSAGE
                    ):
                        self.__close_worker_threads()
                        self.__record_service.exit()
                        message = "Finished recording"
                        logger.warning(message)
                        self.nextion_interface.message = message
                        self.nextion_state.state = NextionPagingState.IDLE

                    elif self.nextion_state.p_state == NextionPagingState.USE_AWAY_MESSAGE:

                        if self.__recorded_message_service.is_running:
                            self.__close_worker_threads()
                            message = "Stopped playing away message..."
                            self.nextion_interface.message = message
                            self.__recorded_message_service.force_exit()
                            logger.warning("CHANGING STATE TO IDLE")
                            self.nextion_state.state = NextionPagingState.IDLE
                        else:
                            self.__close_worker_threads()
                            message = "Finished playing away message"
                            self.nextion_interface.message = message
                            logger.warning("CHANGING STATE TO IDLE")
                            self.nextion_state.state = NextionPagingState.IDLE

                    elif self.nextion_state.p_state == NextionPagingState.SENDING_CUSTOM_MESSAGE:

                        self._page_custom_message(start=False)

                        self.__close_worker_threads()

                        message = "Finished operation"
                        logger.warning(message)
                        self.nextion_interface.message = message
                        if self.__recorded_message_service.is_running:
                            self.__recorded_message_service.force_exit()

                        self.nextion_state.state = NextionPagingState.IDLE

                    # TODO: If a message was automatically answered
                    elif self.nextion_state.p_state == NextionPagingState.UNATTENDED_MODE:
                        self._incoming_call(start=False)
                        self.__close_worker_threads()

                        message = f"Changing state back to UNATTENDED_MODE"
                        logger.info(message)
                        # self.nextion_interface.message = message
                        self._change_nextion_message = True  # This way the Automatic answer message will be displayed once
                        self.nextion_state.state = NextionPagingState.UNATTENDED_MODE
                    # logger.warning("Current running threads after FINISH_OPERATION:")
                    # for thread in enumerate():
                    #     logger.warning(thread.name)

                elif self.nextion_state.state == NextionPagingState.MESSAGE:

                    if not self.__record_service.is_running:
                        message = "Press the Away or Custom message\r\ncheckboxes to start the recording"
                        # logger.info(message)
                        self.nextion_interface.message = message
                    else:
                        "Recording is currently undergoing... Please wait!"
                # TODO
                elif self.nextion_state.state == NextionPagingState.RECORDING_AWAY_MESSAGE:
                    if not self.__record_service.is_running:
                        self.__start_worker_threads(message="Recording Away message...",
                                                    timeout=2)
                        self.__record_service.run("away_message")
                    else:
                        message = "Nothing should happen here. The worker thread is taking" \
                                  "care of the message that needs to be displayed"
                        # logger.info(message)
                        # self.nextion_interface.message = message

                # TODO: Need to add state RECORDING_CUSTOM_MESSAGE = 12 to the low end FW (Talk with Alex)
                elif self.nextion_state.state == NextionPagingState.RECORDING_CUSTOM_MESSAGE:
                    if not self.__record_service.is_running:
                        self.__start_worker_threads(message="Recording Custom message...",
                                                    timeout=2)
                        self.__record_service.run("custom_message")
                    else:
                        message = "Recording Custom message..."
                        # logger.info(message)
                        # self.nextion_interface.message = message

                elif self.nextion_state.state == NextionPagingState.USE_AWAY_MESSAGE:
                    # aplay --channels 1 --format MU_LAW --rate 8000 away_message1.wav

                    if not self.__recorded_message_service.is_running:
                        self.__start_worker_threads(message="Playing away message...", timeout=1)
                        self.__recorded_message_service.play_locally(audio_file_type="away_message", audio_file_id="")

                    elif self.__recorded_message_service.exit():
                        logger.warning("CHANGING STATE TO FINISH OPERATION")
                        self.nextion_state.state = NextionPagingState.FINISH_OPERATION

                elif self.nextion_state.state == NextionPagingState.SENDING_CUSTOM_MESSAGE:
                    # aplay --channels 1 --format MU_LAW --rate 8000 custom_message1.wav

                    # Play locally
                    if not self.__recorded_message_service.is_running:
                        self.__start_worker_threads(message="Playing custom message...", timeout=1)
                        self.__recorded_message_service.play_locally(audio_file_type="custom_message", audio_file_id="")

                    elif self.__recorded_message_service.exit():
                        logger.warning("CHANGING STATE TO FINISH OPERATION")
                        self.nextion_state.state = NextionPagingState.FINISH_OPERATION

                    # Page the message to the selected clients
                    if self.nextion_interface.selected_clients and self.__paging_thread is None:
                        logger.info("<<<<<<<<<<<<<<<< Starting paging >>>>>>>>>>>>>>>>")
                        self._page_custom_message(start=True)
                else:
                    # TODO: implement other states
                    logger.warning(
                        f"State action {self.nextion_state.state} not implemented"
                    )
            # NOTE: The following delay should be carefully selected,
            # some state actions might be missed otherwise.
            time.sleep(0.1)  # time delay between reading each state.
            # Make it small enough such that no state transition is ignored

    def stop(self):
        self.__pmaster.exit()
        self.__stop_services()
