import argparse
import glob
import logging
import logging.config
import os
import pathlib
import subprocess


class RingtoneProcessor:
    """Class to convert a ringtone from mp3 format into PCM_MULAW format wav file."""

    def __init__(self, raw_dir, processed_dir):
        # configure logger
        logging.config.fileConfig("logging.conf")
        self.__logger = logging.getLogger("RingtoneProcessor")

        self.__raw_ringtones = glob.glob(os.path.join(raw_dir, "*.mp3"))
        self.__processed_dir = processed_dir
        # self.processed_ringtones = glob.glob(os.path.join(raw_dir, "*.wav"))
        self.__format = "pcm_mulaw"
        self.__frequency = 8000
        pathlib.Path(self.__processed_dir).mkdir(parents=True, exist_ok=True)

    def _process_ringtones(self):
        """Processes all the mp3 ringtones located in raw_dir using ffmpeg."""
        for raw_ringtone in self.__raw_ringtones:
            self.__logger.info(f"Processing {raw_ringtone}")
            station_name = os.path.splitext(os.path.basename(raw_ringtone))[0]
            command = [
                "ffmpeg",
                "-y",  # to overwrite the output file if exists.
                "-i",
                raw_ringtone,
                "-acodec",
                f"{self.__format}",
                "-ar",
                f"{self.__frequency}",
                "-ac",
                "1",
                os.path.join(self.__processed_dir, f"{station_name}.wav"),
            ]
            process = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            try:
                outs, errs = process.communicate(timeout=2)
                self.__logger.debug(f"output: {outs}")
                self.__logger.debug(f"error :{errs}")
            except subprocess.TimeoutExpired:
                process.kill()
                outs, errs = process.communicate()
                self.__logger.warning(
                    f"Timed out, killing processing of {station_name}.mp3 : errs :{errs}"
                )

    def _delete_processed_files(self):
        """Deletes all the .wav files from the processed ringtones directory.
        Note: Call this method to clear out all older files from the directory.
        """
        processed_ringtones = glob.glob(os.path.join(self.__processed_dir, "*.wav"))
        if len(processed_ringtones) != 0:
            self.__logger.info(f"Deleting processed ringtones")
            for ringtone in processed_ringtones:
                os.remove(ringtone)
                self.__logger.info(f"{ringtone} deleted!")
        else:
            self.__logger.info("No files to process. Folder empty!")

    def _check_processed_ringtones(self):
        """Checks if ringtones in mp3 format have processed counterpart understandable by
        Paging Server's ringtone service.
        """
        processed_ringtones = glob.glob(os.path.join(self.__processed_dir, "*.wav"))
        processed_station_names = set(
            map(lambda x: os.path.splitext(os.path.basename(x))[0], processed_ringtones)
        )
        unprocessed_station_names = set(
            map(
                lambda x: os.path.splitext(os.path.basename(x))[0], self.__raw_ringtones
            )
        )
        if processed_station_names != unprocessed_station_names:
            self.__logger.warning(
                f"Mismatch found in processed and non processed ringtones"
            )
            self.__logger.warning(
                f"MP3 ringtone for stations :{unprocessed_station_names}"
            )
            self.__logger.warning(
                f"MP3 ringtone for stations :{processed_station_names}"
            )

    def run(self):
        """Deletes older .wav files from processed directory and creates processed files from mp3
        ringtones located in raw directory.
        """
        self._delete_processed_files()
        self._process_ringtones()
        self._check_processed_ringtones()


def main():
    parser = argparse.ArgumentParser(
        description="Processes mp3 ringtones to format understood by Paging Server's ringtone service"
    )
    parser.add_argument(
        "--raw_dir",
        help="Path to directory containing mp3 ringtones",
        type=str,
        default="/mnt/data/ringtones",
    )
    parser.add_argument(
        "--proc_dir",
        help="Path to directory containing mp3 ringtones",
        type=str,
        default="/mnt/data/processed_ringtones",
    )
    args = parser.parse_args()
    rprocessor = RingtoneProcessor(args.raw_dir, args.proc_dir)
    rprocessor.run()


if __name__ == "__main__":
    main()
