import pathlib
import time
from datetime import datetime
from os import path, remove
from threading import Thread
import json

import alsaaudio
import audioop
from server.utils import audio_uci_to_config_dict
from utils import get_logger
from utils import read_json
import shared_constants

from nextion.nextion_enum import KeyID
from nextion.nextion_interface import NextionInterface
from nextion.nextion_panel import NextionPanel

from server.services.sound_level_updater import SoundLevelUpdater, get_sound_level

# configure logger
logger = get_logger("MessageRecordService")


class MessageRecordService:
    """Class to control the audio(capture) aspect for recording a custom message
    methods:
        run  : Start the service.
        exit : Stops the service.
    For usage see paging_server's class definition.
    """
    MAX_NEW_FILES = set([str(i) for i in range(1, 5)])

    # TODO: Added nextion_panel: NextionPanel, nextion_interface: NextionInterface
    #  to allow for manual button state manipulation
    def __init__(self, nextion_panel: NextionPanel, nextion_interface: NextionInterface,
                 sound_level_updater: SoundLevelUpdater):
        try:
            self.config = audio_uci_to_config_dict()
        except Exception as e:
            logger.error("Initialization failed")
            raise Exception("MessageRecordService Initialization failed")

        self.__exit_flag = False
        if self.config["audio_format"] == "alaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_A_LAW
        elif self.config["audio_format"] == "ulaw":
            self.__audio_codec = alsaaudio.PCM_FORMAT_MU_LAW
        else:
            raise Exception(
                "Audio codec not recognized by Message Record service, valid option 'ulaw' and 'alaw'."
            )
        self.nextion_panel = nextion_panel
        try:
            self.button_state = nextion_interface.buttons_state  # use for the flip_button function
        except Exception as e:
            logger.warning(f"Error: {e}")
        self._button_id = -1

        self.__record_device = self.config["input_device"]
        self.__frequency = self.config["sample_rate"]
        self.__period_size = 1000  # self.config["sample_rate"]
        self.__num_channel = self.config["num_channels"]

        self.message_type = "default"
        self.current_file = dict.fromkeys(["id", "filename", "sampling_rate", "encoding", "timestamp"])

        self.__message_path = shared_constants.MESSAGE_PATH
        # Initialize directory structure
        pathlib.Path(self.__message_path).mkdir(parents=True, exist_ok=True)
        # Away messages
        self.occupied_away_messages = set([])
        self._away_msg_backup_file = path.join(self.__message_path, shared_constants.AWAY_MESSAGE_BKP_FILE)
        self._away_messages = read_json(self._away_msg_backup_file)
        # Custom messages
        self.occupied_custom_messages = set([])
        self._custom_msg_backup_file = path.join(self.__message_path, shared_constants.CUSTOM_MESSAGE_BKP_FILE)
        self._custom_messages = read_json(self._custom_msg_backup_file)

        # Init SoundLevelUpdater
        self.__sound_level_updater = sound_level_updater  # SoundLevelUpdater()

        # Thread handles
        self.__message_record_thread = None
        self.__blink_button_text_thread = None
        self.__sound_level_updater_thread = None

        self.is_running = False

    def _new_file_name(self) -> None:
        new_id = self.new_id
        if self.message_type == "away_message":
            self.occupied_away_messages.add(new_id)
            self._button_id = KeyID.AWAY_MESSAGE_CHECKBOX
        elif self.message_type == "custom_message":
            self.occupied_custom_messages.add(new_id)
            self._button_id = KeyID.CUSTOM_MESSAGE_CHECKBOX

        file_name = self.message_type + new_id + ".wav"
        full_path = path.join(self.__message_path, file_name)
        logger.info(f"Full path for new file: {full_path}")

        # TODO
        if path.exists(full_path):
            remove(full_path)
            logger.info("Overwriting existing file")
        else:
            logger.info("Creating new file...")

        open(full_path, mode="w").close()  # create new file
        self.current_file = {"id": new_id,
                             "filename": file_name,
                             "timestamp": self._create_timestamp(),
                             "sampling_rate": self.config["sample_rate"],
                             "encoding": self.config["audio_format"]
                             }

    # TODO
    def _already_used_ids(self):
        """
            Use to check the backup files which IDs have been already used
            to prevent unnecessary overwriting of files (e.g. on restart of the app)
        """
        pass

    @property
    def new_file_name(self):
        file = path.join(self.__message_path, self.current_file["filename"])
        return file

    # TODO: Not tested
    @property
    def new_id(self) -> [None, str]:
        """
            Gets an unoccupied new file id.
        """
        available_ids = []
        if self.message_type == "away_message":
            available_ids = MessageRecordService.MAX_NEW_FILES - self.occupied_away_messages
        elif self.message_type == "custom_message":
            available_ids = MessageRecordService.MAX_NEW_FILES - self.occupied_custom_messages
        else:
            # TODO: Take action if the message type is different than expected
            logger.warning("Message type not recognised")
            pass

        # logger.info(f"Available IDs: {available_ids}")

        if len(available_ids) == 0:
            logger.warning("Starting to overwrite files")
            if self.message_type == "away_message":
                self.occupied_away_messages = set([])
                available_ids = MessageRecordService.MAX_NEW_FILES - self.occupied_away_messages
            elif self.message_type == "custom_message":
                self.occupied_custom_messages = set([])
                available_ids = MessageRecordService.MAX_NEW_FILES - self.occupied_custom_messages
            else:
                # TODO: Take action if the message type is different than expected
                logger.warning("Message type not recognised")
                pass

        return min(available_ids)

    @staticmethod
    def _create_timestamp() -> str:
        return str(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    def update_json_backup(self, message_type=""):
        logger.info("Updating json backup file information...")
        if message_type == "away_message":
            try:
                self._away_messages[self.current_file["id"]] = {
                    "filename": self.current_file["filename"],
                    "sampling_rate": self.current_file["sampling_rate"],
                    "encoding": self.current_file["encoding"],
                    "timestamp": self.current_file["timestamp"]
                }
                self.save(self._away_msg_backup_file)
            except Exception as e:
                logger.warning("Couldn't write to the backup file: %s ", self._away_messages)
                logger.warning(f"Error: {e}")

        elif message_type == "custom_message":
            try:
                self._custom_messages[self.current_file["id"]] = {
                    "filename": self.current_file["filename"],
                    "sampling_rate": self.current_file["sampling_rate"],
                    "encoding": self.current_file["encoding"],
                    "timestamp": self.current_file["timestamp"]
                }
                self.save(self._custom_msg_backup_file)
            except Exception as e:
                logger.warning("Couldn't write to the backup file: %s ", self._custom_messages)
                logger.warning(f"Error: {e}")
        else:
            logger.error(f"Unsupported message type: {message_type}")
            raise Exception(f"Message type [{message_type}] not supported!!!")

    def save(self, backup_file):
        with open(backup_file, "w") as f:
            if self.message_type == "away_message":
                json.dump(self._away_messages, f)

            elif self.message_type == "custom_message":
                json.dump(self._custom_messages, f)

        logger.info(f"Backup information was stored to {backup_file}")

    def run(self, message_type):
        if self.__message_record_thread is None:
            self.__message_record_thread = Thread(target=self._record_audio, kwargs={"message_type": message_type})
            self.__message_record_thread.daemon = True
            self.__message_record_thread.setName("Thread-MsgRecord")
            self.__message_record_thread.start()
            # logger.info("Started __message_record_thread : %s", self.__message_record_thread.name)

        if self.__blink_button_text_thread is None:
            self.__blink_button_text_thread = Thread(target=self._blink_button_text)
            self.__blink_button_text_thread.daemon = True
            self.__blink_button_text_thread.setName("Thread-BlinkButtonText")
            self.__blink_button_text_thread.start()
            # logger.info("Started __blink_button_text_thread: %s", self.__blink_button_text_thread.name)

        if self.__sound_level_updater_thread is None:
            self.__sound_level_updater_thread = Thread(target=self.__sound_level_updater.run)
            self.__sound_level_updater_thread.daemon = True
            self.__sound_level_updater_thread.setName("Thread-SoundLvlUpdater")
            self.__sound_level_updater_thread.start()

    def exit(self):
        self.__exit_flag = True

        if self.__sound_level_updater_thread is not None:
            self.__sound_level_updater.exit()
            self.__sound_level_updater_thread.join()
            self.__sound_level_updater_thread = None
            logger.info("Closed Sound level updater thread")

        if self.__blink_button_text_thread is not None:
            self.__blink_button_text_thread.join()
            self.__blink_button_text_thread = None
            logger.info("Closed Blink text thread")

        if self.__message_record_thread is not None:
            self.__message_record_thread.join()
            self.__message_record_thread = None
            logger.info("Closed Recording thread")

        self._reset_buttons_default_state()
        self.__exit_flag = False

    def _get_audio_device(self):
        try:
            # PCM_NONBLOCK PCM_NORMAL
            audio_device = alsaaudio.PCM(
                alsaaudio.PCM_CAPTURE,
                alsaaudio.PCM_NONBLOCK,
                device=self.__record_device,
            )
            audio_device.setperiodsize(self.__period_size)
            audio_device.setrate(self.__frequency)
            audio_device.setchannels(self.__num_channel)
            audio_device.setformat(self.__audio_codec)
        except Exception as e:
            logger.error(f"Failure: Audio device inaccessible: {e}")
            raise Exception("MessageRecordService Failed!")

        return audio_device

    def _blink_button_text(self):
        while self._button_id == -1:
            time.sleep(0.001)
        logger.info("Start blinking")
        logger.info("Button ID: %s", self._button_id)
        while not self.__exit_flag:
            self._sync_text_leds(self._button_id, "flip")
            time.sleep(0.4)

    def _sync_text_leds(self, button, operation):
        import serial
        sync = False
        if operation == "set":
            self.button_state.set_button(button)
            sync = True
        elif operation == "reset":
            self.button_state.reset_button(button)
            sync = True
        elif operation == "flip":
            self.button_state.flip_button(button)
            sync = True

        if sync:
            with serial.Serial(
                    port=self.nextion_panel.serial_port,
                    baudrate=self.nextion_panel.baud_rate,
                    timeout=self.nextion_panel.timeout
            ) as ser:
                self.nextion_panel.sync_leds(ser)

    def _reset_buttons_default_state(self):
        # self._sync_text_leds(KeyID.AWAY_MESSAGE_CHECKBOX, "reset")
        # self._sync_text_leds(KeyID.CUSTOM_MESSAGE_CHECKBOX, "reset")
        self._sync_text_leds(self._button_id, "reset")
        self._button_id = -1  # Reset button_id to default state

    def _record_audio(self, message_type="custom_message"):
        logger.info(f"Start recording message.")
        # When recording audio, have a parameter type of message
        self.message_type = message_type
        self.is_running = True
        audio_input_dev = self._get_audio_device()
        refresh_max_value_counter = 0
        self._new_file_name()

        with open(self.new_file_name, "wb") as message_file:
            while not self.__exit_flag:
                l, data = audio_input_dev.read()
                if l:
                    refresh_max_value_counter += 1
                    if refresh_max_value_counter % 150 == 0:
                        sound_level = audioop.rms(data, 2)

                        # sound_level = audioop.max(data, 2)
                        # sound_level = audioop.avgpp(data, 2)
                        # sound_level = audioop.avg(data, 2)
                        self.__sound_level_updater.set_new_sound_level(direction="input", new_value=sound_level)

                    message_file.write(data)
                time.sleep(0.001)  # NOTE: Prevents excessive CPU usage in case talk is False.

        # Once done writing the data to the file, store it in the json backup
        self.update_json_backup(self.message_type)

        audio_input_dev.close()  # Free the audio resource
        audio_input_dev = None
        self.is_running = False
        logger.info(f"Stop recording message")
