import audioop
import glob
# import logging
# import logging.config
import os
import time
from os import path, remove, listdir, system
import pathlib
from threading import Thread

import alsaaudio
from server.utils import audio_uci_to_config_dict, encode_mp3_to_ulaw_alaw, encode_headerless_files
from server.services.sound_level_updater import SoundLevelUpdater, get_sound_level
from utils import get_logger, get_files
import shared_constants

from barix.web.utils import readFromJSONFile

logger = get_logger("RingService")

# TODO: Have all files in the RAW directory automatically processed depending on the device encoding settings.


class RingService:
    # DEFAULT_RINGTONE_PATH = path.join("/mnt", "data", "icpaging-master", "processed_ringtones", "baring.wav")
    DEFAULT_RINGTONE_PATH = ""  # Set inside _get_processed_ringtones()
    RAW_RINGTONE_DIR_PATH = path.join(shared_constants.PAGING_MASTER_DATA, shared_constants.RAW_RINGTONES)
    RINGTONES_JSON_FILEPATH = path.join(RAW_RINGTONE_DIR_PATH, shared_constants.RINGTONES_JSON)
    PROCESSED_RINGTONE_DIR_PATH = path.join(shared_constants.PAGING_MASTER_DATA, shared_constants.PROCESSED_RINGTONES)
    PERFORM_ENCODING = path.join(shared_constants.PAGING_MASTER_TMP_PATH, shared_constants.ENCODE_FILES_FLAG)
    PAUSE_DURATION = 3  # Time between each ring

    def __init__(self):
        # configure logger
        # logging.config.fileConfig("logging.conf")
        # self.logger = logging.getLogger("RingService")

        try:
            audio_config = audio_uci_to_config_dict()
        except Exception as e:
            logger.error("RingService initialization failed")
            raise Exception("RingService initialization failed")

        self.__exit_flag = False
        self.__audio_format = audio_config["audio_format"]
        if self.__audio_format == "alaw":
            self.__alsa_audio_codec = alsaaudio.PCM_FORMAT_A_LAW
        elif self.__audio_format == "ulaw":
            self.__alsa_audio_codec = alsaaudio.PCM_FORMAT_MU_LAW
        else:
            raise Exception(
                "Audio codec not recognized by Message Record service, valid option 'ulaw' and 'alaw'."
            )
        self.__ringtone_thread = None
        self.__silent_playback = audio_config["silent_playback"]  # default: plug:master1
        self.__playback_device = audio_config["output_device"]  # default: plughw:2,0
        self.__frequency = audio_config["sample_rate"]
        self.__alsa_period_size = audio_config["sample_rate"]
        self.__num_channel = audio_config["num_channels"]
        self.active_station = None
        self.is_running = False
        self.playing_ringtone = False

        self.__output_sound_level_updater = SoundLevelUpdater()
        self.__output_sound_level_updater_thread = None

        # Initialize directory structure. TODO: Remove when reset_defaults() have been integrated
        pathlib.Path(RingService.PROCESSED_RINGTONE_DIR_PATH).mkdir(parents=True, exist_ok=True)
        pathlib.Path(RingService.RAW_RINGTONE_DIR_PATH).mkdir(parents=True, exist_ok=True)
        pathlib.Path(RingService.RINGTONES_JSON_FILEPATH, exist_ok=True)

        # system(f"touch {RingService.PERFORM_ENCODING}")

        if path.exists(RingService.PERFORM_ENCODING):
            remove(RingService.PERFORM_ENCODING)
            logger.info("Encode flag set. Ringtones need to be re-encoded")

            # TODO: When a raw file is deleted (from the webUI), the encoded file remains.
            #  Find a better way to remove it together with the raw file.
            # Remove all encoded files
            for file in get_files(RingService.PROCESSED_RINGTONE_DIR_PATH):
                fullpath = RingService.PROCESSED_RINGTONE_DIR_PATH + "/" + file
                system(f'rm {fullpath}')
                # logger.info(f"Removed {file}")

            # Always process default ringtone first
            if encode_mp3_to_ulaw_alaw(
                    file_id="0_baring.wav",
                    mp3_file_name="0_baring.wav",
                    source_path=RingService.RAW_RINGTONE_DIR_PATH,
                    destination_path=RingService.PROCESSED_RINGTONE_DIR_PATH,
                    station_number='0',
                    encoding_settings=audio_config,
            ):
                logger.info("Default ringtone processed")
            
            if path.exists(RingService.RINGTONES_JSON_FILEPATH):
                ringtones_json_dict = readFromJSONFile(RingService.RINGTONES_JSON_FILEPATH)
                for raw_file in ringtones_json_dict:
                    if encode_mp3_to_ulaw_alaw(
                            file_id=str(raw_file["file_id"]),
                            mp3_file_name=str(raw_file["filename"]),
                            source_path=RingService.RAW_RINGTONE_DIR_PATH,
                            destination_path=RingService.PROCESSED_RINGTONE_DIR_PATH,
                            station_number=str(raw_file["station_number"]),
                            encoding_settings=audio_config,
                    ):
                        pass
                    # elif encode_headerless_files(mp3_file_name=str(file),
                    #                              source_path=RingService.RAW_RINGTONE_DIR_PATH,
                    #                              destination_path=RingService.PROCESSED_RINGTONE_DIR_PATH,
                    #                              station_number="",
                    #                              encoding_settings=audio_config,
                    #                              ):
                    #     logger.info("IT WORKED WITH HEADLESS!")
                    else:
                        logger.warning("Failed to correctly encode ringtone: %s", raw_file)

        self._ringtone_mapping = self._get_processed_ringtones()

    @property
    def ringtone_filename(self):
        if self.active_station is None:
            return RingService.DEFAULT_RINGTONE_PATH
        elif self.active_station in self._ringtone_mapping:
            return path.join(
                RingService.PROCESSED_RINGTONE_DIR_PATH,
                self._ringtone_mapping[self.active_station],
            )
        else:
            logger.info(f"No Ringtone configured for {self.active_station}. Using default.")
            return RingService.DEFAULT_RINGTONE_PATH

    # TODO: The PERFORM_ENCODING flag must be set after audio settings had been changed
    def _get_processed_ringtones(self, use_full_ringtone_name=True):
        """
            Gets all processed ringtone files, and maps them to the corresponding station numbers
            @param use_full_ringtone_name:
                If set to False, the filename has to only contain the station number
            @return:
                Map of station numbers and corresponding ringtone
        """
        if path.exists(RingService.PROCESSED_RINGTONE_DIR_PATH):
            # Only have ringtones for the current sample rate and encoding
            custom_ringtone_paths = glob.glob(
                path.join(RingService.PROCESSED_RINGTONE_DIR_PATH, f"*{self.__audio_format}_{self.__frequency}.wav")
            )
            file_names = map(lambda ringtone: path.basename(ringtone), custom_ringtone_paths)
            file_names = map(lambda file_name: path.splitext(file_name)[0], file_names)
            if use_full_ringtone_name:
                station_names = list(
                    map(lambda file_name: file_name.split('_')[0], file_names)
                )

                ringtone_mapping = {
                    int(station): ringtone  # Use station number at beginning of the ringtones, separated by _
                    for station, ringtone in zip(station_names, custom_ringtone_paths)
                }

            else:
                station_names = list(
                    map(lambda file_name: path.splitext(file_name)[0], file_names)
                )
                ringtone_mapping = {
                    int(station): ringtone  # Use station number at beginning of the ringtones, separated by _
                    for station, ringtone in zip(station_names, custom_ringtone_paths)
                }

        else:
            logger.warning(
                f"No such directory : {RingService.PROCESSED_RINGTONE_DIR_PATH} "
            )
            ringtone_mapping = {}
        logger.info("Ringtone mapping: %s", ringtone_mapping)
        # As 0 is not a valid station name, it is used to mark the default ringtone
        try:
            RingService.DEFAULT_RINGTONE_PATH = ringtone_mapping[0]
        except KeyError as e:
            logger.warning(f"Exception occured: {e}")
            logger.warning("Properly encoded default ringtone was not found!")
            system(f"touch {RingService.PERFORM_ENCODING}")  # Set the encoding flag so the ringtones can be re-encoded
            logger.warning("Please restart the app to automatically correct the encoding.")

        return ringtone_mapping

    def _get_audio_device(self):
        try:
            logger.info(f"HW device to be used for playback: {self.__playback_device}")
            audio_device = alsaaudio.PCM(
                alsaaudio.PCM_PLAYBACK,
                alsaaudio.PCM_NORMAL,
                device=self.__playback_device,  # device=self.__silent_playback, __playback_device
            )
            audio_device.setperiodsize(self.__alsa_period_size)
            audio_device.setrate(self.__frequency)
            audio_device.setchannels(self.__num_channel)
            audio_device.setformat(self.__alsa_audio_codec)
            return audio_device

        except Exception as e:
            logger.error(f"Failure: Audio device inaccessible: {e}")
            raise Exception("RingService Failed!")

    def _play_audio_once(self, audio_output_device):
        sound_level = 0
        with open(self.ringtone_filename, "rb") as f:
            self.playing_ringtone = True
            data = f.read()
            data_block_size = self.__alsa_period_size
            count = 0
            leftover = len(data) % data_block_size

            for i in range(len(data) // data_block_size):
                if not self.__exit_flag:
                    data_block = data[i * data_block_size: (i + 1) * data_block_size]
                    # Using header-less ulaw/alaw files
                    audio_output_device.write(data_block)
                    count = count + data_block_size
                    try:
                        scaling_factor = 1
                        if len(data) % 2 != 0:
                            sound_level = round(audioop.rms(data[:-1], 2) * scaling_factor)
                        else:
                            sound_level = round(audioop.rms(data, 2) * scaling_factor)
                        self.__output_sound_level_updater.set_new_sound_level(direction="output",
                                                                              new_value=sound_level)
                    except Exception as e:
                        logger.warning(f"Following occurred: {e}")
                else:
                    logger.warning("Interrupted rintone")
                    break
            if leftover != 0 and not self.__exit_flag:
                data_block = data[len(data) - leftover:]
                audio_output_device.write(data_block)
                count = count + leftover
            # logger.info(f"Data len: {len(data)}, -> Wrote to alsa: {count}")
        self.playing_ringtone = False

    def _play_audio(self):
        self.is_running = True
        audio_output_dev = self._get_audio_device()
        logger.info("Start ringing")
        count = 0
        delay = 0.01  # Keep the delay short so the ringing can be stopped quickly

        while not self.__exit_flag:
            if count % (RingService.PAUSE_DURATION // delay) == 0:
                logger.info("Ringing...")
                if not self.playing_ringtone:
                    logger.info("Start ring tone...")
                    self._play_audio_once(audio_output_dev)
            count += 1
            time.sleep(delay)
        audio_output_dev.close()
        logger.info("Stop ringing")

    def exit(self):
        self.__exit_flag = True

        if self.__ringtone_thread is not None:
            self.__ringtone_thread.join()
            self.__ringtone_thread = None

        if self.__output_sound_level_updater_thread is not None:
            self.__output_sound_level_updater.exit()
            self.__output_sound_level_updater_thread.join()
            # logger.info("Closed thread Output Sound Level Updater")
            self.__output_sound_level_updater_thread = None

        self.is_running = False
        self.__exit_flag = False

    def run(self):

        # Note: This function is idempotent.
        if self.__ringtone_thread is None:
            self.__ringtone_thread = Thread(target=self._play_audio)
            self.__ringtone_thread.daemon = True
            self.__ringtone_thread.setName("Thread-Ringtone")
            self.__ringtone_thread.start()

        if self.__output_sound_level_updater_thread is None:
            self.__output_sound_level_updater_thread = Thread(target=self.__output_sound_level_updater.run)
            self.__output_sound_level_updater_thread.daemon = True
            self.__output_sound_level_updater_thread.setName("Thread-OutputSoundLvlUpdater")
            # logger.info("Opened thread Output Sound Level Updater")
            self.__output_sound_level_updater_thread.start()
