from time import sleep
import json
import os
from pathlib import Path

from utils import read_json, get_logger
from shared_constants import PAGING_MASTER_TMP_PATH, INPUT_LEVEL, OUTPUT_LEVEL


logger = get_logger(name="SoundLevelUpdater", log_file="/var/log/sound_level_changes")


class SoundLevelUpdater:

    def __init__(self):

        # Make sure the PAGING_MASTER_TMP_PATH exists
        if not os.path.exists(PAGING_MASTER_TMP_PATH):
            os.makedirs(PAGING_MASTER_TMP_PATH)

        self._input_level_file = Path(PAGING_MASTER_TMP_PATH + INPUT_LEVEL)
        self._output_level_file = Path(PAGING_MASTER_TMP_PATH + OUTPUT_LEVEL)

        self._input_level_file.touch(exist_ok=True)  # Create if it doesn't exist
        self._output_level_file.touch(exist_ok=True)  # Create if it doesn't exist

        # Set default levels
        self._record_sound_level_change("input", 0)
        self._record_sound_level_change("output", 0)

        self._direction = "input"
        self._new_value = 0
        self._sound_level_recorded = False
        self.__exit = True

        logger.info("SoundLevelUpdater initialized...")

    def _record_sound_level_change(self, direction, value):
        """
            Writes the current state to the state file
        @return:
        """
        state_dict = dict()
        logger.debug(f"Direction [{direction}], value: {value}")

        state_dict["LEVEL"] = value

        if direction == "output":
            with open(self._output_level_file, "w") as f:
                json.dump(state_dict, f)
                logger.debug(f"State was stored to {self._output_level_file}")
        elif direction == "input":
            with open(self._input_level_file, "w") as f:
                json.dump(state_dict, f)
                logger.debug(f"State was stored to {self._input_level_file}")
        else:
            logger.warning("Unknown sound level direction")

    def set_new_sound_level(self, direction, new_value):
        self._direction = direction
        self._new_value = new_value
        self._sound_level_recorded = False

    def exit(self):
        logger.info("Exiting service")
        self.__exit = True

    def run(self):
        self.__exit = False

        logger.info("Starting service")
        while not self.__exit:
            if not self._sound_level_recorded:
                self._sound_level_recorded = True
                self._record_sound_level_change(self._direction, self._new_value)

            sleep(0.009)  # Should not be divisible by 100ms

        # Reset the sound level value on exit
        logger.debug("Reset sound level to 0")
        self._record_sound_level_change(self._direction, 0)


def get_sound_level(direction):
    sound_level_file = os.path.join(PAGING_MASTER_TMP_PATH, INPUT_LEVEL)
    if direction == "output":
        sound_level_file = os.path.join(PAGING_MASTER_TMP_PATH, OUTPUT_LEVEL)

    try:
        fact_info_json = read_json(sound_level_file)
        device_state = fact_info_json["LEVEL"]
        return device_state
    except Exception as e:
        logger.error("Exception found: {}".format(e))
    return 0
