import os
import threading
import time
from datetime import datetime
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from utils import get_logger, read_json
import shared_constants

logger = get_logger(name="SyncFilesWD", log_file="/mnt/data/icpaging-master/files_wd_log")


class SyncFilesWD:
    """
        Class to detect file changes in a directory and synchronise them between services.
        Param: pattern for files to match in the given directory E.g. "*.wav"
    """

    def __init__(self, pattern: list):
        self.pattern = pattern
        self.path_to_observe = shared_constants.MESSAGE_PATH if len(shared_constants.MESSAGE_PATH) > 1 else '.'

        self.event_handler = PatternMatchingEventHandler(self.pattern, ignore_patterns=None,
                                                         ignore_directories=False, case_sensitive=True)

        self.event_handler.on_created = self.__on_created
        self.event_handler.on_deleted = self.__on_deleted
        self.event_handler.on_modified = self.__on_modified
        self.event_handler.on_moved = self.__on_moved

        self.observer = Observer()
        self.observer.schedule(self.event_handler, self.path_to_observe, recursive=True)

        # Check whether the specified path exists or not
        is_exist = os.path.exists(self.path_to_observe)
        if not is_exist:
            # Create a new directory because it does not exist
            os.makedirs(self.path_to_observe)
            logger.info(f"The new directory {self.path_to_observe} was created!")

        # Away messages
        self._away_msg_backup_file = os.path.join(self.path_to_observe, shared_constants.AWAY_MESSAGE_BKP_FILE)
        self._away_messages = read_json(self._away_msg_backup_file)
        # Custom messages
        self._custom_msg_backup_file = os.path.join(self.path_to_observe, shared_constants.CUSTOM_MESSAGE_BKP_FILE)
        self._custom_messages = read_json(self._custom_msg_backup_file)

        self.sync_flag = False
        self.__exit = False
        self.lock = threading.Lock()

        logger.info(f"Path observed: {self.path_to_observe}")
        # self.synchronise_files_and_backup_info()
        logger.info("SyncFilesWD initialized...")

    @staticmethod
    def __on_created(event):
        logger.info(f"Created {event.src_path}")
        logger.warning("No action currently implemented!")
        time.sleep(0.5)

    def __on_deleted(self, event):
        logger.warning(f"Deleted {event.src_path}")
        # logger.warning("No action currently implemented!")
        # self.sync_flag = True
        # self.synchronise_files_and_backup_info()
        time.sleep(0.5)

    @staticmethod
    def __on_moved(event):
        logger.info(f"Moved {event.src_path} to {event.dest_path}")
        logger.warning("No action currently implemented!")
        time.sleep(0.5)

    def __on_modified(self, event):
        modified_file = event.src_path
        event_list = modified_file.split(self.path_to_observe)
        modified_file = event_list[1]

        logger.info(f"{event.src_path} has been modified")
        logger.info(f"File modified: {modified_file}")
        logger.info("Action has not been implemented")
        time.sleep(3)

    # TODO: Needs implementation and testing
    def synchronise_files_and_backup_info(self):
        if self.sync_flag:
            current_files = self._get_physical_files(self.path_to_observe)
            logger.info(current_files)

            result = self._check_equality(current_files)
            logger.info("Files to be updated: %s", result)

            if len(result) == 0:
                logger.info("No file update needed")
            else:
                logger.info("Backup files need to be updated")
                self._update_json_backup(result)
        else:
            logger.error("Sync not allowed")

    # TODO
    def _update_json_backup(self, file_ids_to_update):
        pass

    # TODO
    def _check_equality(self, physical_files_dict):
        self._away_messages = read_json(self._away_msg_backup_file)
        self._custom_messages = read_json(self._custom_msg_backup_file)

        logger.info("Away: %s", self._away_messages)
        logger.info("Custom: %s", self._custom_messages)

        files_to_be_updated = []

        for file_type in physical_files_dict:

            if file_type == "away_messages":
                # Get the physical file name
                for away_message_file in physical_files_dict[file_type]:
                    logger.info("File:")
                    logger.info(away_message_file)
                    real_file_timestamp = physical_files_dict[file_type][away_message_file]["timestamp"]
                    # Separate the IDs in the backup files
                    for file_id, file_info in self._away_messages.items():
                        logger.info("Backup file info: %s", file_info)
                        # Check the different keys in each ID from the backup files
                        for key in file_info:
                            # logger.info("Current key: %s", key)
                            # logger.info("its value: %s", file_info[key])

                            # Check if the physical filename matches any of the "filename" key values
                            if away_message_file == file_info[key]:
                                logger.info("Matched filenames")

                                utc_time = datetime.strptime(file_info["timestamp"], "%Y-%m-%d %H:%M:%S")
                                old_file_epoch = (utc_time - datetime(1970, 1, 1)).total_seconds()
                                logger.info("Cur: %s", datetime.fromtimestamp(real_file_timestamp).strftime("%Y-%m-%d %H:%M:%S"))
                                logger.info("Old: %s", file_info["timestamp"])

                                timestamp_difference = real_file_timestamp - old_file_epoch
                                logger.info("Difference: %s", timestamp_difference)
                                needs_update = (real_file_timestamp >= old_file_epoch)

                                logger.info("Need update?: ")
                                logger.info(needs_update)
                                if needs_update:
                                    logger.info("%s needs to be updated", file_info[key])
                                    files_to_be_updated.append(file_info[key])
                                    time.sleep(0.001)
                        time.sleep(0.001)
                    time.sleep(0.001)
            elif file_type == "custom_messages":
                pass
        time.sleep(0.001)

        return files_to_be_updated

    # TODO
    def _get_physical_files(self, path, file_extension=".wav"):
        file_dict = dict()
        file_dict["away_messages"] = dict()
        file_dict["custom_messages"] = dict()

        self.sync_flag = False

        for file in os.listdir(path):
            if file.endswith(file_extension):
                filepath = os.path.join(path, file)
                if file.split("_")[0] == "away":
                    file_dict["away_messages"][str(file)] = dict()
                    file_dict["away_messages"][str(file)]["filename"] = str(file)
                    file_dict["away_messages"][str(file)]["filepath"] = str(filepath)
                    file_dict["away_messages"][str(file)]["timestamp"] = round(os.path.getmtime(filepath), 1)

                elif file.split("_")[0] == "custom":
                    file_dict["custom_messages"][str(file)] = dict()
                    file_dict["custom_messages"][str(file)]["filename"] = str(file)
                    file_dict["custom_messages"][str(file)]["filepath"] = str(filepath)
                    file_dict["custom_messages"][str(file)]["timestamp"] = round(os.path.getmtime(filepath), 1)
        return file_dict

    def run(self):
        self.observer.start()
        try:
            while not self.__exit:
                time.sleep(3)
        except KeyboardInterrupt:
            self.observer.stop()
            logger.warning("KeyboardInterrupt. SyncFileWD stopped")
        self.observer.stop()
        self.observer.join()

    def exit(self):
        self.__exit = True



