from utils import get_uci, get_logger
from os import path
import subprocess

logger = get_logger("ServerUtils")

valid_station_numbers = list(range(0, 255))
valid_format_types = ["ulaw", "alaw"]
valid_sample_rates = ["8000", "24000"]


def audio_uci_to_config_dict():
    """Creates config from uci to be used by Audio service of PagingServer"""
    config_dict = {}
    package = "icpaging-master"
    config_dict["paging_priority"] = get_uci(package, "barp", "paging_priority")
    config_dict["input_device"] = get_uci(package, "audio", "input_device")
    config_dict["output_device"] = get_uci(package, "audio", "output_device")
    config_dict["sample_rate"] = int(get_uci(package, "audio", "sample_rate"))
    config_dict["audio_format"] = get_uci(package, "audio", "audio_format")
    config_dict["period_size"] = int(get_uci(package, "audio", "period_size"))
    config_dict["num_channels"] = int(get_uci(package, "audio", "num_channels"))
    config_dict["silent_playback"] = get_uci("application", "audio", "silent_playback")
    return config_dict


def encode_mp3_to_ulaw_alaw(file_id, mp3_file_name, source_path, destination_path,
                            station_number, encoding_settings, is_ringtone=True):

    # source_file = path.join(source_path, mp3_file_name)
    source_file = path.join(source_path, file_id)
    output_audio_format = "mulaw"
    output_codec = "pcm_mulaw"
    if encoding_settings["audio_format"] == "alaw":
        output_audio_format = "alaw"
        output_codec = "pcm_alaw"
    output_smp_rate = encoding_settings["sample_rate"]
    if is_ringtone:
        wav_file_name = create_ringtone_name(station_number=station_number,
                                             raw_file_name=mp3_file_name,
                                             encoding=encoding_settings["audio_format"],
                                             sample_rate=str(output_smp_rate))
    else:
        wav_file_name = path.splitext(path.basename(mp3_file_name))[0] + ".wav"
    destination = path.join(destination_path, wav_file_name)

    # Linux syntax : ffmpeg -y -f mulaw -ar 8000 -ac 1 -i away_message1.wav -ar 24000 away_message1_24.wav
    command = [
        "ffmpeg",
        "-y",  # overwrite the output file if exists.
        "-i",
        source_file,
        "-acodec",
        f"{output_codec}",  # output codec
        "-f",
        f"{output_audio_format}",  # output file format
        "-ar",
        f"{output_smp_rate}",
        "-ac",
        "1",
        destination,
    ]

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        outs, errs = process.communicate(timeout=5)

        if process.returncode == 0:
            logger.info(f"Successful processing of: {mp3_file_name}")
            return True

        # logger.info(f"Outs: {outs}")
        # logger.info(f"Errs: {errs}")
        logger.warning(f"Failed processing of: {mp3_file_name}. Check for header-less source files.")
        return False
    except subprocess.CalledProcessError as e:
        logger.info(f"Errs: {e}")
        logger.warning(f"Failed processing of: {mp3_file_name}")
        return False

    except subprocess.TimeoutExpired:
        process.kill()
        outs, errs = process.communicate()
        logger.warning(
            f"Timed out, killing processing of {mp3_file_name} : errs :{errs}"
        )
        logger.warning("To avoid this, try increasing the timeout so larger files can be processed")
    return False


# Linux syntax:
# ffmpeg -y -f mulaw -ar 8000 -ac 1 -i 0_baring_ulaw_8000.wav -acodec pcm_mulaw -f mulaw -ar 24000 -ac 1 ./test.wav
def encode_headerless_files(mp3_file_name, source_path, destination_path,
                            station_number, encoding_settings, is_ringtone=True):

    source_file = path.join(source_path, mp3_file_name)
    output_audio_format = "mulaw"
    output_codec = "pcm_mulaw"
    if encoding_settings["audio_format"] == "alaw":
        output_audio_format = "alaw"
        output_codec = "pcm_alaw"
    output_smp_rate = encoding_settings["sample_rate"]

    if is_ringtone:
        source_station_n, source_format_type, source_sample_rate = extract_input_file_config(mp3_file_name)

        wav_file_name = create_ringtone_name(station_number=source_station_n if not "" else station_number,
                                             raw_file_name=mp3_file_name,
                                             encoding=encoding_settings["audio_format"],
                                             sample_rate=str(output_smp_rate))
    else:
        wav_file_name = path.splitext(path.basename(mp3_file_name))[0] + ".wav"
    destination = path.join(destination_path, wav_file_name)

    command = [
        "ffmpeg",
        "-y",  # overwrite the output file if exists.
        "-f",
        f"{source_format_type}",  # input file format
        "-ar",
        f"{source_sample_rate}",
        "-i",
        source_file,
        "-acodec",
        f"{output_codec}",  # output codec
        "-f",
        f"{output_audio_format}",  # output file format
        "-ar",
        f"{output_smp_rate}",
        "-ac",
        "1",
        destination,
    ]

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        outs, errs = process.communicate(timeout=5)

        if process.returncode != 0:
            logger.info(f"Outs: {outs}")
            logger.info(f"Errs: {errs}")
            logger.warning(f"Failed processing of: {mp3_file_name}")
            return False
        logger.info(f"Successful processing header-less source: {mp3_file_name}")
        return True
    except subprocess.CalledProcessError as e:
        logger.info(f"Errs: {e}")
        logger.warning(f"Failed processing of: {mp3_file_name}")
        return False
    except subprocess.TimeoutExpired:
        process.kill()
        logger.warning(
            f"Timed out, killing processing of {mp3_file_name}"
        )
        logger.warning("To avoid this, try increasing the timeout so larger files can be processed")
    return False


def create_ringtone_name(station_number, raw_file_name, encoding, sample_rate):
    basename = path.splitext(path.basename(raw_file_name))[0]
    basename_components = basename.split("_")
    extension = ".wav"
    st_number = station_number

    new_name_components_list = list()

    # logger.info("Basename components: %s", basename_components)

    if basename_components[0] == "0":
        # This is a default tone
        st_number = "0"

    # Remove the settings from the raw filename
    for component in basename_components:
        if component not in str(valid_station_numbers) and \
            component not in valid_format_types and \
                component not in valid_sample_rates:
            new_name_components_list.append(component)

    basename = "_".join(new_name_components_list)
    # Add the new settings to the new filename
    name_components = [st_number, basename, encoding, sample_rate]
    new_name = "_".join(name_components) + extension

    # logger.info("New filename created: %s", new_name)

    return new_name


# TODO: Implementation not finished. Do not use this function yet
def strip_file_settings(file_path: str = '', file_extension: str = '') -> bool:
    new_name_components_list = list()

    filename_components = path.splitext(path.basename(file_path))
    basename = filename_components[0]
    basename_components = basename.split("_")
    extension = file_extension if file_extension != '' else filename_components[1]

    # Remove the settings from the raw filename
    for component in basename_components:
        if component not in str(valid_station_numbers) and \
                component not in valid_format_types and \
                component not in valid_sample_rates:
            new_name_components_list.append(component)

    return False


def extract_input_file_config(file_name: str):
    """
        A ringtone filename components are:
        file_name[0]: station number
        file_name[1]: ringtone name
        file_name[2]: format (ulaw or alaw)
        file_name[3]: sample rate (8000 or 24000)
    @param file_name: Name of the headerless ringtone file
    @return: station_number, encoding, sample_rate
    """
    # Default values
    station_number = ""
    format_type = ""
    sample_rate = ""

    name_components_list = path.splitext(path.basename(file_name))[0].split("_")
    # logger.info("Name components list: %s", name_components_list)

    if int(name_components_list[0]) in valid_station_numbers:
        station_number = name_components_list[0]
    if name_components_list[2] in valid_format_types:
        format_type = "mulaw" if name_components_list[2] == "ulaw" else "alaw"
    if name_components_list[3] in valid_sample_rates:
        sample_rate = name_components_list[3]

    return station_number, format_type, sample_rate
