import json
from subprocess import PIPE, Popen, TimeoutExpired  # Needed to sent commands to the linux terminal
from pathlib import Path
from os import path, system, listdir

import logging
import logging.config

from shared_constants import (PAGING_MASTER_DATA, RAW_RINGTONES, PROCESSED_RINGTONES,
                              PAGING_MASTER_SOURCE_BASE, RAW_DEFAULT_RINGTONE, GPIO_PATH)


def get_uci(package, section, option):
    command = ["uci", "get", f"{package}.{section}.{option}"]
    output, error = Popen(command, stdout=PIPE, stderr=PIPE).communicate()
    if len(error) == 0:
        return output.decode().strip("\n")
    else:
        error_message = (
            f"Unable to get {package}.{section}.{option} uci setting : {error}"
        )
        raise Exception(error_message)


def read_json(file_name):
    if not Path(file_name).exists():
        Path(file_name).touch(exist_ok=True)
        # Create the file if it doesn't exist
        with open(Path(file_name), "w") as f:
            f.write("{}")

    with open(file_name, "r") as f:
        return json.load(f)


def get_files(filepath):
    """
        Yields one by one all files in a directory.
    @param filepath: Full directory to where the files are located
    @return: Name of file
    """
    for file in listdir(filepath):
        if path.isfile(path.join(filepath, file)):
            yield file


def get_logger(name="", config="", log_file="/var/log/messages"):
    if config == "":
        print(f"Creating default logger for {name}")
        new_logger = logging.getLogger(name)
        new_logger.setLevel(logging.INFO)
        # Initialize directory structure
        try:
            Path(PAGING_MASTER_DATA).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(f"Directory {PAGING_MASTER_DATA} already exists")
        log_file = Path(log_file, exist_ok=True)
        ch = logging.FileHandler(log_file)
        ch.setLevel(logging.INFO)

        # create formatter, to get path: - [%(pathname)s]
        formatter = logging.Formatter(
            "%(asctime)s barix-ipam400 " + name + "[%(lineno)d] - %(levelname)s - %(message)s [%(threadName)s]",
            datefmt="%b %d %H:%M:%S",
        )

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        new_logger.addHandler(ch)
    else:
        print("About to load .conf file: %s", config)
        logging.config.fileConfig(config)
        new_logger = logging.getLogger(name)
        print("Loaded the .conf file")
    return new_logger


logger = get_logger("MainUtils")


# /sys/class/gpio# echo 1 > gpio14/value
def mcu_gpio_control(gpio_name: str, gpio_state: bool):
    gpio = GPIO_PATH + gpio_name + "/value"

    if gpio_state:
        value = 1
    else:
        value = 0

    command = [
        '/bin/echo',
        f"{value}",
        ">",
        gpio
    ]
    try:
        command_str = " ".join(command)
        system(command_str)
        logger.debug("Command executed: %s", command_str)
        return True

    except TimeoutExpired as e:
        logger.error(
            f"Timed out while setting gpio {gpio_name}: error :{e}"
        )
        return False


def validate_directory_structure():
    """
        Use this function to ensure the file structure has been successfully created
    @return:
    """
    # TODO
    raw_ringtone_dir = path.join(PAGING_MASTER_DATA, RAW_RINGTONES)
    proc_ringtone_dir = path.join(PAGING_MASTER_DATA, PROCESSED_RINGTONES)

    try:
        # Initialize directory structure
        if not Path(PAGING_MASTER_DATA).exists():
            reset_defaults()
        else:
            Path(raw_ringtone_dir).mkdir(parents=True, exist_ok=True)
            Path(proc_ringtone_dir).mkdir(parents=True, exist_ok=True)
        return True
    except Exception as e:
        raise FileNotFoundError(e)


def reset_defaults():
    logger.info("Reset to defaults flag has been set. Executing reset_defaults()")

    raw_ringtone_dir = path.join(PAGING_MASTER_DATA, RAW_RINGTONES)
    proc_ringtone_dir = path.join(PAGING_MASTER_DATA, PROCESSED_RINGTONES)
    raw_default_ringtone = path.join(PAGING_MASTER_SOURCE_BASE, RAW_DEFAULT_RINGTONE)

    # Initialize directory structure
    Path(PAGING_MASTER_DATA).mkdir(parents=True, exist_ok=True)
    Path(raw_ringtone_dir).mkdir(parents=True, exist_ok=True)
    Path(proc_ringtone_dir).mkdir(parents=True, exist_ok=True)

    # Add different commands to be executed to this list, as a list of strings
    commands_list = [
        ["cp", f"{raw_default_ringtone}", f"{raw_ringtone_dir}", ],
    ]

    for command in commands_list:
        result = execute_command(command)
        if result:
            logger.info("Success")
        else:
            logger.info("Failed executing command: %s", " ".join(command))


def restart_main_app(system_func: bool = True):
    # This function is currently used when the Nextion panel is changing the number of pages
    command = [
        '/etc/init.d/icpaging-master',
        "restart",
    ]

    if system_func:
        system(" ".join(command))
        return

    return execute_command(command)


def load_alsa_settings(system_func: bool = True):
    logger.info("Loading alsa settings")
    command = [
        'alsactl',
        'restore',
    ]
    if system_func:
        system(" ".join(command))
        return
    return execute_command(command)


def store_alsa_settings(system_func: bool = True):
    logger.info("Storing alsa settings")
    command = [
        'alsactl',
        'store',
    ]
    if system_func:
        system(" ".join(command))
        return
    return execute_command(command)


def execute_command(command: list):
    """
        Open a process to execute the given command
    @param command: List of strings
    @return: True if successful, False otherwise
    """

    process = Popen(
        command,
        stdout=PIPE,
        stderr=PIPE,
    )
    try:
        # logger.info("Executing command: %s", " ".join(command))
        outs, errs = process.communicate(timeout=5)

        if process.returncode != 0:
            logger.error(f"Error occurred: {errs}")
            return False
        # logger.info("Success")
        return True

    except Exception as e:
        logger.warning(f"While trying to execute command: [{command}], the following exception occurred: {e}")
        return False


# ---------------------------------- Custom exceptions -------------------------------
class NextionPanelReset(RuntimeError):
    def __init__(self, message="Main app needs to be restarted!"):
        self.message = message
        super().__init__(self.message)
